#include "Accelerate.h"

#include <stdlib.h>
#include <math.h>

#include "resize.h"
#include "warp.h"

#include "dispatch_parallel.h"

// Internal util functions

static inline vImage_Error vImageValidateBuffer(const vImage_Buffer *buffer) {
    if (!buffer) return kvImageNullPointerArgument;

    if (!buffer->data || buffer->rowBytes < buffer->width) return kvImageInvalidParameter;

    return kvImageNoError;
}

static inline vImage_Error vImageValidateBuffers(
    const vImage_Buffer *src,
    const vImage_Buffer *dst)
{
    vImage_Error error = vImageValidateBuffer(src);
    if (error) return error;

    error = vImageValidateBuffer(dst);
    if (error) return error;

    if (src->width != dst->width || src->height != dst->height) return kvImageBufferSizeMismatch;

    return kvImageNoError;
}

static inline vImage_Error vImageValidateBuffersROI(
    const vImage_Buffer *src,
    const vImage_Buffer *dst,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y)
{
    vImage_Error error = vImageValidateBuffer(src);
    if (error) return error;

    error = vImageValidateBuffer(dst);
    if (error) return error;

    if (srcOffsetToROI_X > src->width) return kvImageInvalidOffset_X;

    if (srcOffsetToROI_Y > src->height) return kvImageInvalidOffset_Y;

    if ((srcOffsetToROI_X + dst->width) > src->width
            || (srcOffsetToROI_Y + dst->height) > src->height) return kvImageRoiLargerThanInputBuffer;

    return kvImageNoError;
}

static inline int min(int a, int b) {
    return b < a ? b : a;
}

static inline int max(int a, int b) {
    return b > a ? b : a;
}

static inline Pixel_8 saturate_cast_to_Pixel_8(int v) {
    if ((unsigned int)v <= 255)
        return v;
    if (v < 0)
        return 0;
    return 255;
}



// Blending

typedef struct AlphaBlendContext_ {
    const vImage_Buffer *srcTop;
    const vImage_Buffer *srcTopAlpha;
    const vImage_Buffer *srcBottom;
    const vImage_Buffer *dest;
    Pixel_8 constAlpha;
} AlphaBlendContext;

static void parallel_vImagePremultipliedConstAlphaBlend_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const AlphaBlendContext* cnt = (AlphaBlendContext*)context;
    const vImagePixelCount w = cnt->srcTop->width;

    const Pixel_8* top = GET_CONST_LINE_8(cnt->srcTop, y);
    const Pixel_8* bottom = GET_CONST_LINE_8(cnt->srcBottom, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    const Pixel_8 alpha = cnt->constAlpha;
    const Pixel_8 oma = 255 - alpha;

    for (int x = 0; x < w; ++x, top += 4, bottom += 4, dest += 4) {
        dest[0] = (top[0] * alpha + bottom[0] * oma) / 255;
        dest[1] = (top[1] * alpha + bottom[1] * oma) / 255;
        dest[2] = (top[2] * alpha + bottom[2] * oma) / 255;
        dest[3] = (top[3] * alpha + bottom[3] * oma) / 255;
    }
}

vImage_Error vImagePremultipliedConstAlphaBlend_ARGB8888(
    const vImage_Buffer *srcTop,
    Pixel_8 constAlpha,
    const vImage_Buffer *srcBottom,
    const vImage_Buffer *dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(srcTop, srcBottom);
    if (error) return error;

    error = vImageValidateBuffers(srcTop, dest);
    if (error) return error;

    const vImagePixelCount h = srcTop->height;

    AlphaBlendContext context;
    context.srcTop = srcTop;
    context.srcBottom = srcBottom;
    context.dest = dest;
    context.constAlpha = constAlpha;

    dispatch_parallel(parallel_vImagePremultipliedConstAlphaBlend_ARGB8888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageAlphaBlend_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const AlphaBlendContext* cnt = (AlphaBlendContext*)context;
    const vImagePixelCount w = cnt->srcTop->width;

    const Pixel_8* top = GET_CONST_LINE_8(cnt->srcTop, y);
    const Pixel_8* bottom = GET_CONST_LINE_8(cnt->srcBottom, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    for (int x = 0; x < w; ++x, top += 4, bottom += 4, dest += 4) {
        const Pixel_8 topAlpha = top[0];
        const Pixel_8 bottomAlpha = bottom[0];
        const Pixel_8 resultAlpha = (topAlpha * 255 + (255 - topAlpha) * bottomAlpha + 127) / 255;
        if (resultAlpha == 0) {
            dest[0] = 0;
            dest[1] = 0;
            dest[2] = 0;
            dest[3] = 0;
            continue;
        }
        const Pixel_8 bottomCoef = ((255 - topAlpha) * bottomAlpha + 127) / 255;

        dest[0] = resultAlpha;
        dest[1] = (topAlpha * top[1] + bottomCoef * bottom[1] +  127) / resultAlpha;
        dest[2] = (topAlpha * top[2] + bottomCoef * bottom[2] +  127) / resultAlpha;
        dest[3] = (topAlpha * top[3] + bottomCoef * bottom[3] +  127) / resultAlpha;
    }
}

vImage_Error vImageAlphaBlend_ARGB8888(
    const vImage_Buffer *srcTop,
    const vImage_Buffer *srcBottom,
    const vImage_Buffer *dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(srcTop, srcBottom);
    if (error) return error;

    error = vImageValidateBuffers(srcTop, dest);
    if (error) return error;

    const vImagePixelCount h = srcTop->height;

    AlphaBlendContext context;
    context.srcTop = srcTop;
    context.srcBottom = srcBottom;
    context.dest = dest;

    dispatch_parallel(parallel_vImageAlphaBlend_ARGB8888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageAlphaBlend_NonpremultipliedToPremultiplied_Planar8(void* context, size_t iter) {
    const size_t y = iter;
    const AlphaBlendContext* cnt = (AlphaBlendContext*)context;
    const vImagePixelCount w = cnt->srcTop->width;

    const Pixel_8* top = GET_CONST_LINE_8(cnt->srcTop, y);
    const Pixel_8* top_alpha = GET_CONST_LINE_8(cnt->srcTopAlpha, y);
    const Pixel_8* bottom = GET_CONST_LINE_8(cnt->srcBottom, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    for (int x = 0; x < w; ++x, ++top, ++top_alpha, ++bottom, ++dest) {
    	*dest = ( (*top_alpha) * (*top) + (255 - (*top_alpha) ) * (*bottom) ) / 255;
    }
}

vImage_Error vImageAlphaBlend_NonpremultipliedToPremultiplied_Planar8(
    const vImage_Buffer *srcTop,
    const vImage_Buffer *srcTopAlpha,
    const vImage_Buffer *srcBottom,
    const vImage_Buffer *dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(srcTop, srcBottom);
    if (error) return error;

    error = vImageValidateBuffers(srcTop, dest);
    if (error) return error;

    const vImagePixelCount h = srcTop->height;

    AlphaBlendContext context;
    context.srcTop = srcTop;
    context.srcTopAlpha = srcTopAlpha;
    context.srcBottom = srcBottom;
    context.dest = dest;

    dispatch_parallel(parallel_vImageAlphaBlend_NonpremultipliedToPremultiplied_Planar8, (int)h, &context);

    return kvImageNoError;
}



// Overwrite channels

typedef struct OverwriteChannelsContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    const vImage_Buffer *newSrc;
    Pixel_8 scalar8;
    Pixel_F scalarF;
    uint32_t masked_pixel;
    uint32_t mask;
    uint32_t inv_mask;
    const uint8_t *permuteMap;
} OverwriteChannelsContext;

static void parallel_vImageOverwriteChannelsWithScalar_Planar8(void* context, size_t iter) {
    const size_t y = iter;
    const OverwriteChannelsContext* cnt = (OverwriteChannelsContext*)context;
    memset(GET_LINE_8(cnt->dest, y), cnt->scalar8, cnt->dest->width);
}

vImage_Error vImageOverwriteChannelsWithScalar_Planar8(
    Pixel_8 scalar,
    const vImage_Buffer *dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffer(dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    OverwriteChannelsContext context;
    context.dest = dest;
    context.scalar8 = scalar;

    dispatch_parallel(parallel_vImageOverwriteChannelsWithScalar_Planar8, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageOverwriteChannelsWithScalar_PlanarF(void* context, size_t iter) {
    const size_t y = iter;
    const OverwriteChannelsContext* cnt = (OverwriteChannelsContext*)context;

    const vImagePixelCount w = cnt->dest->width;
    Pixel_F* dest = GET_LINE_F(cnt->dest, y);
    const Pixel_F scalar = cnt->scalarF;

    for (int x = 0; x < w; ++x, ++dest)
        *dest = scalar;
}

vImage_Error vImageOverwriteChannelsWithScalar_PlanarF(
    Pixel_F scalar,
    const vImage_Buffer *dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffer(dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    OverwriteChannelsContext context;
    context.dest = dest;
    context.scalarF = scalar;

    dispatch_parallel(parallel_vImageOverwriteChannelsWithScalar_PlanarF, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageOverwriteChannelsWithPixel_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const OverwriteChannelsContext* cnt = (OverwriteChannelsContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const uint32_t* src = (const uint32_t*)GET_CONST_LINE_8(cnt->src, y);
    uint32_t* dest = (uint32_t*)GET_LINE_8(cnt->dest, y);

    const uint32_t masked_pixel = cnt->masked_pixel;
    const uint32_t inv_mask = cnt->inv_mask;

    for (int x = 0; x < w; ++x, ++src, ++dest)
        *dest = masked_pixel | ((*src) & inv_mask);
}

vImage_Error vImageOverwriteChannelsWithPixel_ARGB8888(
    const Pixel_8888 the_pixel,
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    uint8_t copyMask,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    uint32_t pixel = 0;
    memcpy(&pixel, &the_pixel[0], 4);

    uint32_t mask = 0;
    Pixel_8* pmask = (Pixel_8*)&mask;
    if (copyMask & 0x8)
        pmask[0] = 255;
    if (copyMask & 0x4)
        pmask[1] = 255;
    if (copyMask & 0x2)
        pmask[2] = 255;
    if (copyMask & 0x1)
        pmask[3] = 255;

    uint32_t inv_mask = ~mask;

    OverwriteChannelsContext context;
    context.src = src;
    context.dest = dest;
    context.masked_pixel = pixel & mask;
    context.inv_mask = inv_mask;

    dispatch_parallel(parallel_vImageOverwriteChannelsWithPixel_ARGB8888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageOverwriteChannels_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const OverwriteChannelsContext* cnt = (OverwriteChannelsContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const uint32_t* src = (const uint32_t*)GET_CONST_LINE_8(cnt->src, y);
    uint32_t* dest = (uint32_t*)GET_LINE_8(cnt->dest, y);
    const Pixel_8* newSrc = GET_CONST_LINE_8(cnt->newSrc, y);
    uint32_t pixel = 0;
    Pixel_8* ppixel = (Pixel_8*)&pixel;

    const uint32_t mask = cnt->mask;
    const uint32_t inv_mask = cnt->inv_mask;

    for (int x = 0; x < w; ++x, ++src, ++newSrc, ++dest) {
        ppixel[3] = ppixel[2] = ppixel[1] = ppixel[0] = *newSrc;
        *dest = (pixel & mask) | ((*src) & inv_mask);
    }
}

vImage_Error vImageOverwriteChannels_ARGB8888(
    const vImage_Buffer *newSrc,
    const vImage_Buffer *origSrc,
    const vImage_Buffer *dest,
    uint8_t copyMask,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(newSrc, origSrc);
    if (error) return error;

    error = vImageValidateBuffers(newSrc, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    uint32_t mask = 0;
    Pixel_8* pmask = (Pixel_8*)&mask;
    if (copyMask & 0x8)
        pmask[0] = 255;
    if (copyMask & 0x4)
        pmask[1] = 255;
    if (copyMask & 0x2)
        pmask[2] = 255;
    if (copyMask & 0x1)
        pmask[3] = 255;

    uint32_t inv_mask = ~mask;

    OverwriteChannelsContext context;
    context.src = origSrc;
    context.dest = dest;
    context.newSrc = newSrc;
    context.mask = mask;
    context.inv_mask = inv_mask;

    dispatch_parallel(parallel_vImageOverwriteChannels_ARGB8888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImagePermuteChannels_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const OverwriteChannelsContext* cnt = (OverwriteChannelsContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_8* dst = GET_LINE_8(cnt->dest, y);

    const uint8_t pm_0 = cnt->permuteMap[0];
    const uint8_t pm_1 = cnt->permuteMap[1];
    const uint8_t pm_2 = cnt->permuteMap[2];
    const uint8_t pm_3 = cnt->permuteMap[3];

    for (int x = 0; x < w; ++x, src += 4, dst += 4) {
        const Pixel_8 dst_0 = src[pm_0];
        const Pixel_8 dst_1 = src[pm_1];
        const Pixel_8 dst_2 = src[pm_2];
        const Pixel_8 dst_3 = src[pm_3];
        dst[0] = dst_0;
        dst[1] = dst_1;
        dst[2] = dst_2;
        dst[3] = dst_3;
    }
}

vImage_Error vImagePermuteChannels_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const uint8_t permuteMap[4],
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    OverwriteChannelsContext context;
    context.src = src;
    context.dest = dest;
    context.permuteMap = permuteMap;

    dispatch_parallel(parallel_vImagePermuteChannels_ARGB8888, (int)h, &context);

    return kvImageNoError;
}



// Lookup table

typedef struct LookupTableContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    const Pixel_8 *table;
    const Pixel_8 *alphaTable;
    const Pixel_8 *redTable;
    const Pixel_8 *greenTable;
    const Pixel_8 *blueTable;
} LookupTableContext;

static void parallel_vImageTableLookUp_Planar8(void* context, size_t iter) {
    const size_t y = iter;
    const LookupTableContext* cnt = (LookupTableContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);
    const Pixel_8* table = cnt->table;

    for (int x = 0; x < w; ++x, ++src, ++dest)
        *dest = table[*src];
}

vImage_Error vImageTableLookUp_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const Pixel_8 table[256],
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    LookupTableContext context;
    context.src = src;
    context.dest = dest;
    context.table = table;

    dispatch_parallel(parallel_vImageTableLookUp_Planar8, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageTableLookUp_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const LookupTableContext* cnt = (LookupTableContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);
    const Pixel_8* alphaTable = cnt->alphaTable;
    const Pixel_8* redTable = cnt->redTable;
    const Pixel_8* greenTable = cnt->greenTable;
    const Pixel_8* blueTable = cnt->blueTable;

    for (int x = 0; x < w; ++x, src += 4, dest += 4) {
        dest[0] = alphaTable[src[0]];
        dest[1] = redTable  [src[1]];
        dest[2] = greenTable[src[2]];
        dest[3] = blueTable [src[3]];
    }
}

vImage_Error vImageTableLookUp_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const Pixel_8 alphaTable[256],
    const Pixel_8 redTable[256],
    const Pixel_8 greenTable[256],
    const Pixel_8 blueTable[256],
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    static const Pixel_8 copyTable[256] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                                            11, 12, 13, 14, 15, 16, 17, 18, 19,
                                            20, 21, 22, 23, 24, 25, 26, 27, 28,
                                            29, 30, 31, 32, 33, 34, 35, 36, 37,
                                            38, 39, 40, 41, 42, 43, 44, 45, 46,
                                            47, 48, 49, 50, 51, 52, 53, 54, 55,
                                            56, 57, 58, 59, 60, 61, 62, 63, 64,
                                            65, 66, 67, 68, 69, 70, 71, 72, 73,
                                            74, 75, 76, 77, 78, 79, 80, 81, 82,
                                            83, 84, 85, 86, 87, 88, 89, 90, 91,
                                            92, 93, 94, 95, 96, 97, 98, 99, 100,
                                            101, 102, 103, 104, 105, 106, 107,
                                            108, 109, 110, 111, 112, 113, 114,
                                            115, 116, 117, 118, 119, 120, 121,
                                            122, 123, 124, 125, 126, 127, 128,
                                            129, 130, 131, 132, 133, 134, 135,
                                            136, 137, 138, 139, 140, 141, 142,
                                            143, 144, 145, 146, 147, 148, 149,
                                            150, 151, 152, 153, 154, 155, 156,
                                            157, 158, 159, 160, 161, 162, 163,
                                            164, 165, 166, 167, 168, 169, 170,
                                            171, 172, 173, 174, 175, 176, 177,
                                            178, 179, 180, 181, 182, 183, 184,
                                            185, 186, 187, 188, 189, 190, 191,
                                            192, 193, 194, 195, 196, 197, 198,
                                            199, 200, 201, 202, 203, 204, 205,
                                            206, 207, 208, 209, 210, 211, 212,
                                            213, 214, 215, 216, 217, 218, 219,
                                            220, 221, 222, 223, 224, 225, 226,
                                            227, 228, 229, 230, 231, 232, 233,
                                            234, 235, 236, 237, 238, 239, 240,
                                            241, 242, 243, 244, 245, 246, 247,
                                            248, 249, 250, 251, 252, 253, 254,
                                            255 };

    LookupTableContext context;
    context.src = src;
    context.dest = dest;
    context.alphaTable = alphaTable != NULL ? alphaTable : copyTable;
    context.redTable   = redTable != NULL   ? redTable   : copyTable;
    context.greenTable = greenTable != NULL ? greenTable : copyTable;
    context.blueTable  = blueTable != NULL  ? blueTable  : copyTable;

    dispatch_parallel(parallel_vImageTableLookUp_ARGB8888, (int)h, &context);

    return kvImageNoError;
}



// Scale

vImage_Error vImageScale_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffer(src);
    if (error) return error;
    error = vImageValidateBuffer(dest);
    if (error) return error;

    resize_8888(
        (const unsigned char*)src->data,
        src->width,
        src->height,
        src->rowBytes,
        (unsigned char*)dest->data,
        dest->width,
        dest->height,
        dest->rowBytes);

    return kvImageNoError;
}

vImage_Error vImageScale_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffer(src);
    if (error) return error;
    error = vImageValidateBuffer(dest);
    if (error) return error;

    resize_8(
        (const unsigned char*)src->data,
        src->width,
        src->height,
        src->rowBytes,
        (unsigned char*)dest->data,
        dest->width,
        dest->height,
        dest->rowBytes);

    return kvImageNoError;
}

vImage_Error vImageScale_PlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffer(src);
    if (error) return error;
    error = vImageValidateBuffer(dest);
    if (error) return error;

    resize_F(
        (const float*)src->data,
        src->width,
        src->height,
        src->rowBytes,
        (float*)dest->data,
        dest->width,
        dest->height,
        dest->rowBytes);

    return kvImageNoError;
}

// Histogram calculating

typedef struct HistogramContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    Pixel_8 min_val;
    Pixel_8 diff;
} HistogramContext;

vImage_Error vImageHistogramCalculation_Planar8(
    const vImage_Buffer *src,
    vImagePixelCount *histogram,
    vImage_Flags flags)
{
    vImagePixelCount width = src->width;
    vImagePixelCount height = src->height;
    memset(histogram, 0, 256 * sizeof(vImagePixelCount));

    for(vImagePixelCount y = 0; y < height; ++y) {
        const Pixel_8 * sp = (const Pixel_8 *)src->data + y * src->rowBytes;
        for(vImagePixelCount x = 0; x < width; ++x, ++sp)
            ++histogram[*sp];
    }

    return kvImageNoError;
}

vImage_Error vImageHistogramCalculation_ARGB8888(
    const vImage_Buffer *src,
    vImagePixelCount *histogram[4],
    vImage_Flags flags)
{
    vImagePixelCount width = src->width;
    vImagePixelCount height = src->height;
    memset(histogram[0], 0, 256 * sizeof(vImagePixelCount));
    memset(histogram[1], 0, 256 * sizeof(vImagePixelCount));
    memset(histogram[2], 0, 256 * sizeof(vImagePixelCount));
    memset(histogram[3], 0, 256 * sizeof(vImagePixelCount));

    for(vImagePixelCount y = 0; y < height; ++y) {
        const Pixel_8 * sp = (const Pixel_8 *)src->data + y * src->rowBytes;
        for(vImagePixelCount x = 0; x < width; ++x, sp += 4) {
            ++histogram[0][sp[0]];
            ++histogram[1][sp[1]];
            ++histogram[2][sp[2]];
            ++histogram[3][sp[3]];
        }
    }

    return kvImageNoError;
}

static void parallel_vImageContrastStretch_Planar8(void* context, size_t iter) {
    const size_t y = iter;
    const HistogramContext* cnt = (HistogramContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    const Pixel_8 min_val = cnt->min_val;
    const Pixel_8 diff = cnt->diff;

    for (int x = 0; x < w; ++x, ++src, ++dest)
        *dest = (*src - min_val) * 255 / diff;
}

vImage_Error vImageContrastStretch_Planar8 (
   const vImage_Buffer *src,
   const vImage_Buffer *dest,
   vImage_Flags flags)
{
    vImagePixelCount w = src->width;
    vImagePixelCount h = src->height;

    Pixel_8 max_val = 0, min_val = 255;

    for (vImagePixelCount y = 0; y < h; ++y) {
        const Pixel_8* sp = GET_CONST_LINE_8(src, y);
        for (vImagePixelCount x = 0; x < w; ++x, ++sp) {
            if (*sp > max_val) max_val = *sp;
            if (*sp < min_val) min_val = *sp;
        }
    }

    const Pixel_8 diff = max_val - min_val;

    if (diff > 0) {
        HistogramContext context;
        context.src = src;
        context.dest = dest;
        context.min_val = min_val;
        context.diff = diff;

        dispatch_parallel(parallel_vImageContrastStretch_Planar8, (int)h, &context);
    }
    else if (src->data != dest->data) {
        for (vImagePixelCount y = 0; y < h; ++y) {
            const Pixel_8* sp = GET_CONST_LINE_8(src, y);
            Pixel_8* dp = GET_LINE_8(dest, y);
            memcpy(dp, sp, w);
        }
    }

    return kvImageNoError;
}


// Format conversion

typedef struct FormatConversionContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    Pixel_8 alpha;
    const vImage_Buffer *red;
    const vImage_Buffer *green;
    const vImage_Buffer *blue;
    Pixel_F k;
    Pixel_F minF;
} FormatConversionContext;

static void parallel_vImageConvert_ARGB8888toRGB888(void* context, size_t iter) {
    const size_t y = iter;
    const FormatConversionContext* cnt = (FormatConversionContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    for (int x = 0; x < w; ++x, src += 4, dest += 3) {
        dest[0] = src[1];
        dest[1] = src[2];
        dest[2] = src[3];
    }
}

vImage_Error vImageConvert_ARGB8888toRGB888(
    const vImage_Buffer *argbSrc,
    const vImage_Buffer *rgbDest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(argbSrc, rgbDest);
    if (error) return error;

    const vImagePixelCount h = rgbDest->height;

    FormatConversionContext context;
    context.src = argbSrc;
    context.dest = rgbDest;

    dispatch_parallel(parallel_vImageConvert_ARGB8888toRGB888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageConvert_Planar8ToXRGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const FormatConversionContext* cnt = (FormatConversionContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8 alpha = cnt->alpha;
    const Pixel_8* red = GET_CONST_LINE_8(cnt->red, y);
    const Pixel_8* green = GET_CONST_LINE_8(cnt->green, y);
    const Pixel_8* blue = GET_CONST_LINE_8(cnt->blue, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    for (int x = 0; x < w; ++x, ++red, ++green, ++blue, dest += 4) {
        dest[0] = alpha;
        dest[1] = *red;
        dest[2] = *green;
        dest[3] = *blue;
    }
}

vImage_Error vImageConvert_Planar8ToXRGB8888(
    Pixel_8 alpha,
    const vImage_Buffer *red,
    const vImage_Buffer *green,
    const vImage_Buffer *blue,
    const vImage_Buffer *dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(dest, red);
    if (error) return error;
    error = vImageValidateBuffers(dest, green);
    if (error) return error;
    error = vImageValidateBuffers(dest, blue);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    FormatConversionContext context;
    context.alpha = alpha;
    context.red = red;
    context.green = green;
    context.blue = blue;
    context.dest = dest;

    dispatch_parallel(parallel_vImageConvert_Planar8ToXRGB8888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImageConvert_Planar8toPlanarF(void* context, size_t iter) {
    const size_t y = iter;
    const FormatConversionContext* cnt = (FormatConversionContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_F* dest = GET_LINE_F(cnt->dest, y);
    const Pixel_F k = cnt->k;
    const Pixel_F minF = cnt->minF;

    for (int x = 0; x < w; ++x, ++src, ++dest)
        *dest = (*src) * k + minF;
}

vImage_Error vImageConvert_Planar8toPlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    Pixel_F maxFloat,
    Pixel_F minFloat,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    FormatConversionContext context;
    context.src = src;
    context.dest = dest;
    context.k = (maxFloat - minFloat) / 255.f;
    context.minF = minFloat;

    dispatch_parallel(parallel_vImageConvert_Planar8toPlanarF, (int)h, &context);

    return kvImageNoError;
}



// Convolve

typedef struct ConvolveContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    const float * kernel;
    vImagePixelCount kernel_h;
    vImagePixelCount kernel_w;
    const vImagePixelCount *integral;
} ConvolveContext;

void parallel_vImageBoxConvolve_Planar8(void* context, size_t iter) {
    const size_t y = iter;
    const ConvolveContext* cnt = (ConvolveContext*) context;

    const vImagePixelCount width = cnt->dest->width;
    const vImagePixelCount height = cnt->dest->height;

    const vImagePixelCount kernel_w = cnt->kernel_w;
    const vImagePixelCount kernel_h = cnt->kernel_h;

    const vImagePixelCount * integral = cnt->integral;
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);

    const vImagePixelCount dx = kernel_w / 2;
    const vImagePixelCount dy = kernel_h / 2;

    for (vImagePixelCount x = 0; x < width; ++x, ++dest) {
        const vImagePixelCount left = max((int)x - dx, 0);
        const vImagePixelCount right = min((int)x + dx, width - 1);
        const vImagePixelCount up = max((int)y - dy, 0);
        const vImagePixelCount down = min((int)y + dy, height - 1);
        const vImagePixelCount S = (right - left + 1) * (down - up + 1);

        const vImagePixelCount int_A = *(integral + down * width + right);
        const vImagePixelCount int_B = left == 0 ? 0 : *(integral + down * width + (left - 1));
        const vImagePixelCount int_C = up == 0 ? 0 : *(integral + (up - 1) * width + right);
        const vImagePixelCount int_D = (left == 0 || up == 0) ? 0 : *(integral + (up - 1) * width + (left - 1));

        *dest = (int_A + int_D + S / 2 - int_B - int_C) / S;
    }
}

vImage_Error vImageBoxConvolve_Planar8(
   const vImage_Buffer *src,
   const vImage_Buffer *dest,
   void *tempBuffer,
   vImagePixelCount srcOffsetToROI_X,
   vImagePixelCount srcOffsetToROI_Y,
   uint32_t kernel_height,
   uint32_t kernel_width,
   Pixel_8 backgroundColor,
   vImage_Flags flags)
{
    if (kernel_width % 2 == 0 || kernel_height % 2 == 0) return kvImageInvalidKernelSize;

    vImage_Error error = vImageValidateBuffersROI(src, dest, srcOffsetToROI_X, srcOffsetToROI_Y);
    if (error) return error;

    vImage_Rect roi = { srcOffsetToROI_X, srcOffsetToROI_Y, dest->width, dest->height };
    vImage_Buffer src_roi = GET_SUB_IMAGE(src, roi, GET_LINE_8);

    vImagePixelCount width = src_roi.width;
    vImagePixelCount height = src_roi.height;

    Pixel_8 * img = (Pixel_8 *) src_roi.data;

    // calculate integral image.
    vImagePixelCount* integral = (vImagePixelCount*)calloc(width * height, sizeof(vImagePixelCount));

    integral[0] = img[0];

    vImagePixelCount rb = src_roi.rowBytes;

    for (vImagePixelCount y = 1; y < height; ++y) {
        vImagePixelCount diff = y * width;
        *(integral + diff) = *(img + y * rb) + *(integral + diff - width);
    }

    for (vImagePixelCount x = 1; x < width; ++x)
        *(integral + x) = *(img + x) + *(integral + x - 1);

    for (vImagePixelCount y = 1; y < height; ++y) {
        Pixel_8 * img_current = img + y * rb + 1;
        vImagePixelCount * int_current = integral + y * width + 1;
        for (vImagePixelCount x = 1; x < width;
                ++x, ++int_current, ++img_current) {
            *int_current = *(img_current) + *(int_current - 1)
                    + *(int_current - width) - *(int_current - width - 1);
        }
    }

    ConvolveContext context;
    context.dest = (vImage_Buffer *) dest;
    context.integral = integral;
    context.kernel_h = kernel_height;
    context.kernel_w = kernel_width;

    dispatch_parallel(parallel_vImageBoxConvolve_Planar8, (int)height, &context);

    free(integral);

    return kvImageNoError;
}

void parallel_vImageBoxConvolve_ARGB8888(void * context, size_t iter)
{
    const size_t y = iter;
    const ConvolveContext* cnt = (ConvolveContext*) context;

    const vImagePixelCount width = cnt->dest->width;
    const vImagePixelCount height = cnt->dest->height;

    const vImagePixelCount kernel_w = cnt->kernel_w;
    const vImagePixelCount kernel_h = cnt->kernel_h;

    const vImagePixelCount * integral = cnt->integral;
    Pixel_8* dest = (Pixel_8 *) GET_LINE_8(cnt->dest, y);

    const vImagePixelCount dx = kernel_w / 2;
    const vImagePixelCount dy = kernel_h / 2;

    const vImagePixelCount rb_int = 4 * width;

    for (int x = 0; x < width; ++x, dest += 4) {
        const vImagePixelCount left = max((int)x - dx, 0);
        const vImagePixelCount right = min((int)x + dx, width - 1);
        const vImagePixelCount up = max((int)y - dy, 0);
        const vImagePixelCount down = min((int)y + dy, height - 1);
        const vImagePixelCount S = (right - left + 1) * (down - up + 1);

        dest[0] = 255;

        for(int i = 1; i < 4; ++i) {
            const int int_A = (integral + down * rb_int + right * 4)[i];
            const int int_B = left == 0 ? 0 : (integral + down * rb_int + (left - 1) * 4)[i];
            const int int_C = up == 0 ? 0 : (integral + (up - 1) * rb_int + right * 4)[i];
            const int int_D = (left == 0 || up == 0) ? 0 : (integral + (up - 1) * rb_int + (left - 1) * 4)[i];
            dest[i] = (int_A - int_B + int_D - int_C) / S;
        }
    }
}

vImage_Error vImageBoxConvolve_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_8888 backgroundColor,
    vImage_Flags flags)
{
    vImage_Error error = kvImageNoError;
    error = vImageValidateBuffersROI(src, dest, srcOffsetToROI_X, srcOffsetToROI_Y);
    if (error) return error;

    vImage_Rect roi = {srcOffsetToROI_X, srcOffsetToROI_Y, dest->width, dest->height};
    vImage_Buffer src_roi = GET_SUB_IMAGE(src, roi, GET_LINE_8);

    vImagePixelCount width = src_roi.width;
    vImagePixelCount height = src_roi.height;

    Pixel_8 * img = (Pixel_8 *) src_roi.data;

    // calculate integral image.
    vImagePixelCount* integral = (vImagePixelCount*)calloc(4 * width * height, sizeof(vImagePixelCount));

    integral[1] = img[1];
    integral[2] = img[2];
    integral[3] = img[3];

    vImagePixelCount rb_img = src_roi.rowBytes;
    vImagePixelCount rb_int = 4 * width;

    vImagePixelCount * cur_int = integral + rb_int;
    Pixel_8 * cur_img = img + rb_img;

    for (vImagePixelCount y = 1; y < height; ++y, cur_int += rb_int, cur_img += rb_img) {
        cur_int[1] = cur_img[1] + (cur_int - rb_int)[1];
        cur_int[2] = cur_img[2] + (cur_int - rb_int)[2];
        cur_int[3] = cur_img[3] + (cur_int - rb_int)[3];
    }

    cur_int = integral + 4;
    cur_img = img + 4;

    for (vImagePixelCount x = 1; x < width; ++x, cur_int += 4, cur_img += 4) {
        cur_int[1] = cur_img[1] + (cur_int - 4)[1];
        cur_int[2] = cur_img[2] + (cur_int - 4)[2];
        cur_int[3] = cur_img[3] + (cur_int - 4)[3];
    }

    for (vImagePixelCount y = 1; y < height; ++y) {
        cur_img = img + y * rb_img + 4;
        cur_int = integral + y * rb_int + 4;
        for (vImagePixelCount x = 1; x < width; ++x, cur_int += 4, cur_img += 4) {
            cur_int[1] = cur_img[1] + (cur_int - 4)[1] + (cur_int - rb_int)[1] - (cur_int - rb_int - 4)[1];
            cur_int[2] = cur_img[2] + (cur_int - 4)[2] + (cur_int - rb_int)[2] - (cur_int - rb_int - 4)[2];
            cur_int[3] = cur_img[3] + (cur_int - 4)[3] + (cur_int - rb_int)[3] - (cur_int - rb_int - 4)[3];
        }
    }

    ConvolveContext context;
    context.dest = (vImage_Buffer *) dest;
    context.integral = integral;
    context.kernel_h = kernel_height;
    context.kernel_w = kernel_width;

    dispatch_parallel(parallel_vImageBoxConvolve_ARGB8888, (int)height, &context);

    free(integral);

    return kvImageNoError;
}

vImage_Error vImageTentConvolve_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_8 backgroundColor,
    vImage_Flags flags)
{
    if (kernel_width % 2 == 0 || kernel_height % 2 == 0)
        return kvImageInvalidKernelSize;

    vImage_Error error = vImageValidateBuffersROI(src, dest, srcOffsetToROI_X, srcOffsetToROI_Y);
    if (error) return error;

    int kernel_height_0 = (kernel_height + 1) / 2;
    int kernel_height_1 = kernel_height_0;
    if (kernel_height_0 % 2 == 0) {
        kernel_height_0 += 1;
        kernel_height_1 -= 1;
    }

    int kernel_width_0 = (kernel_width + 1) / 2;
    int kernel_width_1 = kernel_width_0;
    if (kernel_width_0 % 2 == 0) {
        kernel_width_0 += 1;
        kernel_width_1 -= 1;
    }

    vImagePixelCount w = dest->width;
    vImagePixelCount h = dest->height;
    vImage_Buffer temp = {malloc(h*w*Pixel_8_sz), h, w, w*Pixel_8_sz};

    error = vImageBoxConvolve_Planar8(src, &temp, tempBuffer, srcOffsetToROI_X, srcOffsetToROI_Y, kernel_height_0, kernel_width_0, backgroundColor, flags);
    if (error) {
        free(temp.data);
        return error;
    }

    error = vImageBoxConvolve_Planar8(&temp, dest, tempBuffer, 0, 0, kernel_height_1, kernel_width_1, backgroundColor, flags);
    free(temp.data);

    return error;
}

vImage_Error vImageTentConvolve_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_8888 backgroundColor,
    vImage_Flags flags)
{
    if (kernel_width % 2 == 0 || kernel_height % 2 == 0)
        return kvImageInvalidKernelSize;

    vImage_Error error = vImageValidateBuffersROI(src, dest, srcOffsetToROI_X, srcOffsetToROI_Y);
    if (error) return error;

    int kernel_height_0 = (kernel_height + 1) / 2;
    int kernel_height_1 = kernel_height_0;
    if (kernel_height_0 % 2 == 0) {
        kernel_height_0 += 1;
        kernel_height_1 -= 1;
    }

    int kernel_width_0 = (kernel_width + 1) / 2;
    int kernel_width_1 = kernel_width_0;
    if (kernel_width_0 % 2 == 0) {
        kernel_width_0 += 1;
        kernel_width_1 -= 1;
    }

    vImagePixelCount w = dest->width;
    vImagePixelCount h = dest->height;
    vImage_Buffer temp = {malloc(h*w*Pixel_8888_sz), h, w, w*Pixel_8888_sz};

    error = vImageBoxConvolve_ARGB8888(src, &temp, tempBuffer, srcOffsetToROI_X, srcOffsetToROI_Y, kernel_height_0, kernel_width_0, backgroundColor, flags);
    if (error) {
        free(temp.data);
        return error;
    }

    error = vImageBoxConvolve_ARGB8888(&temp, dest, tempBuffer, 0, 0, kernel_height_1, kernel_width_1, backgroundColor, flags);
    free(temp.data);

    return error;
}

void parallel_vImageConvolve_PlanarF(void * context, size_t iter)
{
    const size_t y = iter;
    const ConvolveContext* cnt = (ConvolveContext*) context;

    const vImage_Buffer * src = cnt->src;

    const vImagePixelCount width = src->width;
    const vImagePixelCount height = src->height;

    Pixel_F * dest = (Pixel_F *) GET_LINE_F(cnt->dest, y);

    const float * kernel = (float *) cnt->kernel;
    const vImagePixelCount kernel_w = cnt->kernel_w;
    const vImagePixelCount kernel_h = cnt->kernel_h;
    const int dx = kernel_w / 2;
    const int dy = kernel_h / 2;

    for (vImagePixelCount x = 0; x < width; ++x, ++dest) {
        *dest = 0;
        for (int py = -dy; py <= dy; ++py) {
            int cy = y + py;
            cy = (cy < 0) ? 0 : (cy >= height) ? height - 1 : cy;
            const Pixel_F * color_str = (Pixel_F *) GET_LINE_F(src, cy);
            const float * kernel_str = (float *) (kernel + (py + dy) * kernel_w);
            for (int px = -dx; px <= dx; ++px) {
                int cx = x + px;
                cx = (cx < 0) ? 0 : (cx >= width) ? width - 1 : cx;
                *dest += kernel_str[px + dx] * color_str[cx];
            }
        }
    }
}

vImage_Error vImageConvolve_PlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    const float *kernel,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_F backgroundColor,
    vImage_Flags flags
)
{
    vImage_Error error = kvImageNoError;
    error = vImageValidateBuffersROI(src, dest, srcOffsetToROI_X, srcOffsetToROI_Y);
    if (error) return error;

    vImage_Rect roi = {srcOffsetToROI_X, srcOffsetToROI_Y, dest->width, dest->height};
    vImage_Buffer src_roi = GET_SUB_IMAGE(src, roi, GET_LINE_F);

    vImagePixelCount height = src_roi.height;

    ConvolveContext context;
    context.src = (vImage_Buffer *) &src_roi;
    context.dest = dest;
    context.kernel_h = kernel_height;
    context.kernel_w = kernel_width;
    context.kernel = kernel;

    dispatch_parallel(parallel_vImageConvolve_PlanarF, (int)height, &context);

    return kvImageNoError;
}

void parallel_vImageConvolve_ARGBFFFF(void * context, size_t iter)
{
    const size_t y = iter;
    const ConvolveContext* cnt = (ConvolveContext*) context;

    const vImage_Buffer * src = cnt->src;

    const vImagePixelCount width = src->width;
    const vImagePixelCount height = src->height;

    Pixel_F * dest = (Pixel_F *) GET_LINE_F(cnt->dest, y);

    const float * kernel = (float *) cnt->kernel;
    const vImagePixelCount kernel_w = cnt->kernel_w;
    const vImagePixelCount kernel_h = cnt->kernel_h;
    const int dx = kernel_w / 2;
    const int dy = kernel_h / 2;

    for (vImagePixelCount x = 0; x < width; ++x, dest += 4) {
        for (int i = 0; i < 4; ++i)
            dest[i] = 0;
        for (int py = -dy; py <= dy; ++py) {
            int cy = y + py;
            cy = (cy < 0) ? 0 : (cy >= height) ? height - 1 : cy;
            const Pixel_F * color_str = (Pixel_F *) GET_LINE_F(src, cy);
            const float * kernel_str = (float *) (kernel + (py + dy) * kernel_w);
            for (int px = -dx; px <= dx; ++px) {
                int cx = x + px;
                cx = (cx < 0) ? 0 : (cx >= width) ? width - 1 : cx;
                const float kernel_coeff = kernel_str[px + dx];
                const Pixel_F * color = color_str + cx * 4;
                for (int i = 0; i < 4; ++i)
                    dest[i] += kernel_coeff * color[i];
            }
        }
    }
}

vImage_Error vImageConvolve_ARGBFFFF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    const float *kernel,
    uint32_t kernel_height,
    uint32_t kernel_width,
    const Pixel_FFFF backgroundColor,
    vImage_Flags flags
)
{
    vImage_Error error = kvImageNoError;
    error = vImageValidateBuffersROI(src, dest, srcOffsetToROI_X, srcOffsetToROI_Y);
    if (error) return error;

    vImage_Rect roi = {srcOffsetToROI_X, srcOffsetToROI_Y, dest->width, dest->height};
    vImage_Buffer src_roi = GET_SUB_IMAGE(src, roi, GET_LINE_F);

    vImagePixelCount height = src_roi.height;

    ConvolveContext context;
    context.src = (vImage_Buffer *) &src_roi;
    context.dest = dest;
    context.kernel_h = kernel_height;
    context.kernel_w = kernel_width;
    context.kernel = kernel;

    dispatch_parallel(parallel_vImageConvolve_ARGBFFFF, (int)height, &context);

    return kvImageNoError;
}

// Matrix multiply

typedef struct MatrixMultiplyContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    const int16_t *matrix;
    int32_t divisor;
    const int16_t *pre_bias;
    const int32_t *post_bias;
} MatrixMultiplyContext;

static void parallel_vImageMatrixMultiply_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const MatrixMultiplyContext* cnt = (MatrixMultiplyContext*)context;

    const vImagePixelCount w = cnt->dest->width;

    const Pixel_8* src = GET_CONST_LINE_8(cnt->src, y);
    Pixel_8* dest = GET_LINE_8(cnt->dest, y);
    const int16_t *matrix = cnt->matrix;
    const int32_t divisor = cnt->divisor;
    const int16_t *pre_bias = cnt->pre_bias;
    const int32_t *post_bias = cnt->post_bias;

    const int16_t aa = matrix[ 0];
    const int16_t ar = matrix[ 1];
    const int16_t ag = matrix[ 2];
    const int16_t ab = matrix[ 3];

    const int16_t ra = matrix[ 4];
    const int16_t rr = matrix[ 5];
    const int16_t rg = matrix[ 6];
    const int16_t rb = matrix[ 7];

    const int16_t ga = matrix[ 8];
    const int16_t gr = matrix[ 9];
    const int16_t gg = matrix[10];
    const int16_t gb = matrix[11];

    const int16_t ba = matrix[12];
    const int16_t br = matrix[13];
    const int16_t bg = matrix[14];
    const int16_t bb = matrix[15];

    const int16_t pre_a = pre_bias[0];
    const int16_t pre_r = pre_bias[1];
    const int16_t pre_g = pre_bias[2];
    const int16_t pre_b = pre_bias[3];

    const int32_t post_a = post_bias[0];
    const int32_t post_r = post_bias[1];
    const int32_t post_g = post_bias[2];
    const int32_t post_b = post_bias[3];

    for (int x = 0; x < w; ++x, src += 4, dest += 4) {
        const int16_t a = src[0] + pre_a;
        const int16_t r = src[1] + pre_r;
        const int16_t g = src[2] + pre_g;
        const int16_t b = src[3] + pre_b;
        dest[0] = saturate_cast_to_Pixel_8((aa * a + ra * r + ga * g + ba * b + post_a) / divisor);
        dest[1] = saturate_cast_to_Pixel_8((ar * a + rr * r + gr * g + br * b + post_r) / divisor);
        dest[2] = saturate_cast_to_Pixel_8((ag * a + rg * r + gg * g + bg * b + post_g) / divisor);
        dest[3] = saturate_cast_to_Pixel_8((ab * a + rb * r + gb * g + bb * b + post_b) / divisor);
    }
}

vImage_Error vImageMatrixMultiply_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const int16_t matrix[16],
    int32_t divisor,
    const int16_t *pre_bias,
    const int32_t *post_bias,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    int16_t pre_zero[] = {0, 0, 0, 0};
    int32_t post_zero[] = {0, 0, 0, 0};

    MatrixMultiplyContext context;
    context.src = src;
    context.dest = dest;
    context.matrix = matrix;
    context.divisor = divisor;
    context.pre_bias = pre_bias != NULL ? pre_bias : pre_zero;
    context.post_bias = post_bias != NULL ? post_bias : post_zero;

    dispatch_parallel(parallel_vImageMatrixMultiply_ARGB8888, (int)h, &context);

    return kvImageNoError;
}



// Clip

typedef struct ClipContext_ {
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    Pixel_F maxF;
    Pixel_F minF;
} ClipContext;

void parallel_vImageClip_PlanarF(void* context, size_t iter) {
    const size_t y = iter;
    const ClipContext* cnt = (ClipContext*) context;
    const vImagePixelCount w = cnt->dest->width;

    const Pixel_F* src = GET_CONST_LINE_F(cnt->src, y);
    Pixel_F* dest = GET_LINE_F(cnt->dest, y);

    const Pixel_F maxF = cnt->maxF;
    const Pixel_F minF = cnt->minF;

    for (int x = 0; x < w; ++x, ++src, ++dest) {
        if (*dest < minF)
            *dest = minF;
        else if (*dest > maxF)
            *dest = maxF;
    }
}

vImage_Error vImageClip_PlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    Pixel_F maxFloat,
    Pixel_F minFloat,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(src, dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    ClipContext context;
    context.src = src;
    context.dest = dest;
    context.maxF = maxFloat;
    context.minF = minFloat;

    dispatch_parallel(parallel_vImageClip_PlanarF, (int)h, &context);

    return kvImageNoError;
}

// Rotate
typedef struct RotateContext_
{
    const vImage_Buffer *src;
    const vImage_Buffer *dest;
    uint8_t rotationConstant;
} RotateContext;

void parallel_vImageRotate90_Planar8(void * context, size_t iter)
{
    const size_t y = iter;
    const RotateContext* cnt = (RotateContext*) context;

    uint8_t rotationConstant = cnt->rotationConstant;
    const vImagePixelCount w = cnt->dest->width;
    const vImagePixelCount sw = cnt->src->width;
    const vImagePixelCount sh = cnt->src->height;

    const Pixel_8* src = (const Pixel_8*)cnt->src->data;
    Pixel_8 * dest = GET_LINE_8(cnt->dest, y);
    const vImagePixelCount rb = cnt->src->rowBytes;

    switch (rotationConstant) {
        case 0: {
            Pixel_8 * base = (Pixel_8 *)src + y * rb;
            memcpy(dest, base, w);
            break;
        }
        case 1: {
            Pixel_8 * base = (Pixel_8 *)src + (sh - 1) * rb + y;
            for (size_t x = 0; x < w; ++x, base -= rb, dest += 1) {
                *dest = *base;
            }
            break;
        }
        case 2: {
            Pixel_8 * base = (Pixel_8 *)src + (sh - 1 - y) * rb + (sw - 1);
            for (size_t x = 0; x < w; ++x, base -= 1, dest += 1) {
                *dest = *base;
            }
            break;
        }
        case 3: {
            Pixel_8 * base = (Pixel_8 *)src + (sw - 1 - y);
            for (size_t x = 0; x < w; ++x, base += rb, dest += 1) {
                *dest = *base;
            }
            break;
        }
    }
}

vImage_Error vImageRotate90_Planar8(
   const vImage_Buffer *src,
   const vImage_Buffer *dest,
   uint8_t rotationConstant,
   Pixel_8 backColor,
   vImage_Flags flags)
{
    vImagePixelCount sh = src->height;
    vImagePixelCount sw = src->width;
    vImagePixelCount dh = dest->height;
    vImagePixelCount dw = dest->width;
    if (!(((sh == dh) && (sw == dw) && (rotationConstant % 2 == 0))
            || ((sh == dw) && (sw == dh) && (rotationConstant % 2 == 1)))) return kvImageBufferSizeMismatch;

    RotateContext context;
    context.src = src;
    context.dest = dest;
    context.rotationConstant = rotationConstant;

    dispatch_parallel(parallel_vImageRotate90_Planar8, (int)dh, &context);

    return kvImageNoError;
}

void parallel_vImageRotate90_ARGB8888(void * context, size_t iter)
{
    const size_t y = iter;
    const RotateContext* cnt = (RotateContext*) context;

    uint8_t rotationConstant = cnt->rotationConstant;
    const vImagePixelCount w = cnt->dest->width;
    const vImagePixelCount sw = cnt->src->width;
    const vImagePixelCount sh = cnt->src->height;

    const Pixel_8* src = (const Pixel_8*)cnt->src->data;
    Pixel_8 * dest = GET_LINE_8(cnt->dest, y);
    const vImagePixelCount rb = cnt->src->rowBytes;

    switch (rotationConstant) {
        case 0: {
            Pixel_8 * base = (Pixel_8 *)src + y * rb;
            memcpy(dest, base, w*4);
            break;
        }
        case 1: {
            Pixel_8 * base = (Pixel_8 *)src + (sh - 1) * rb + y * 4;
            for (size_t x = 0; x < w; ++x, base -= rb, dest += 4) {
                dest[0] = base[0];
                dest[1] = base[1];
                dest[2] = base[2];
                dest[3] = base[3];
            }
            break;
        }
        case 2: {
            Pixel_8 * base = (Pixel_8 *)src + (sh - 1 - y) * rb + (sw - 1) * 4;
            for (size_t x = 0; x < w; ++x, base -= 4, dest += 4) {
                dest[0] = base[0];
                dest[1] = base[1];
                dest[2] = base[2];
                dest[3] = base[3];
            }
            break;
        }
        case 3: {
            Pixel_8 * base = (Pixel_8 *)src + (sw - y - 1) * 4;
            for (size_t x = 0; x < w; ++x, base += rb, dest += 4) {
                dest[0] = base[0];
                dest[1] = base[1];
                dest[2] = base[2];
                dest[3] = base[3];
            }
            break;
        }
    }
}

vImage_Error vImageRotate90_ARGB8888(
   const vImage_Buffer *src,
   const vImage_Buffer *dest,
   uint8_t rotationConstant,
   Pixel_8888 backColor,
   vImage_Flags flags)
{
    vImagePixelCount sh = src->height;
    vImagePixelCount sw = src->width;
    vImagePixelCount dh = dest->height;
    vImagePixelCount dw = dest->width;
    if (!(((sh == dh) && (sw == dw) && (rotationConstant % 2 == 0))
            || ((sh == dw) && (sw == dh) && (rotationConstant % 2 == 1)))) return kvImageBufferSizeMismatch;

    RotateContext context;
    context.src = src;
    context.dest = dest;
    context.rotationConstant = rotationConstant;

    dispatch_parallel(parallel_vImageRotate90_ARGB8888, (int)dh, &context);

    return kvImageNoError;
}

vImage_Error vImageRotate_Planar8 (
    const vImage_Buffer *src,
    const vImage_Buffer *dst,
    void *tempBuffer,
    float angle,
    Pixel_8 backColor,
    vImage_Flags flags)
{
    // create angle from 0 to 2*pi
    angle = fmod(angle, (float)(2 * M_PI));
    if(angle < 0.f)  angle += 2 * M_PI;

    // centers of dst and src
    const Pixel_D dst_x = (Pixel_D)dst->width  / 2.0,
                  dst_y = (Pixel_D)dst->height / 2.0,
                  src_x = (Pixel_D)src->width  / 2.0,
                  src_y = (Pixel_D)src->height / 2.0;

    // create rotation matrix
    Pixel_D M[6];
    {
        Pixel_D _cos = cos(angle);
        Pixel_D _sin = sin(angle);
        M[0] =  _cos;
        M[1] =  _sin;
        M[2] = -_cos * src_x - _sin * src_y + dst_x;
        M[3] = -_sin;
        M[4] =  _cos;
        M[5] =  _sin * src_x - _cos * src_y + dst_y;
    }

    warpAffine_Planar8(src, dst, M, flags, backColor);
    return kvImageNoError;
}

vImage_Error vImageRotate_ARGB8888 (
    const vImage_Buffer *src,
    vImage_Buffer *dst,
    void *tempBuffer,
    float angle,
    Pixel_8888 backColor,
    vImage_Flags flags)
{
    // create angle from 0 to 2*pi
    angle = fmod(angle, (float)(2 * M_PI));
    if(angle < 0.f)  angle += 2 * M_PI;

    // centers of dst and src
    const Pixel_D dst_x = (Pixel_D)dst->width  / 2.0,
                  dst_y = (Pixel_D)dst->height / 2.0,
                  src_x = (Pixel_D)src->width  / 2.0,
                  src_y = (Pixel_D)src->height / 2.0;

    // create rotation matrix
    Pixel_D M[6];
    {
        Pixel_D _cos = cos(angle);
        Pixel_D _sin  = sin(angle);
        M[0] =  _cos;
        M[1] = _sin;
        M[2] = -_cos * src_x - _sin * src_y + dst_x;
        M[3] = -_sin;
        M[4] = _cos;
        M[5] =  _sin * src_x - _cos * src_y + dst_y;
    }

    warpAffine_ARGB8888(src, dst, M, flags, backColor);
    return kvImageNoError;

}

typedef struct HorizontalReflectContext_ {
    const vImage_Buffer *source;
    const vImage_Buffer *dest;
} HorizontalReflectContext;

static void parallel_vImageHorizontalReflect_ARGB8888(void* context, size_t iter) {
	const size_t y = iter;
	const HorizontalReflectContext* cnt = (HorizontalReflectContext*)context;
	const int w = (int)cnt->source->width;

	const Pixel_8* src = GET_CONST_LINE_8(cnt->source, y);
	Pixel_8* dst = GET_LINE_8(cnt->dest, y);

	for (int x = 0; x < w; ++x) {
		int xl = 4*x;
		int xr = 4*(w - x -1);

		dst[xl]   = src[xr]  ;
		dst[xl+1] = src[xr+1];
		dst[xl+2] = src[xr+2];
		dst[xl+3] = src[xr+3];
	}
}

vImage_Error vImageHorizontalReflect_ARGB8888(
    const vImage_Buffer* source,
    vImage_Buffer* dest,
    vImage_Flags flags)
{
    vImage_Error error = vImageValidateBuffers(source, dest);
    if (error) return error;

    const vImagePixelCount h = source->height;

    HorizontalReflectContext context;
    context.source = source;
    context.dest = dest;

    dispatch_parallel(parallel_vImageHorizontalReflect_ARGB8888, (int)h, &context);

    return kvImageNoError;
}

typedef struct PremuliplyDataContext_ {
    const vImage_Buffer *source;
    const vImage_Buffer *dest;
} PremuliplyDataContext;

static void parallel_vImagePremultiplyData_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const PremuliplyDataContext* cnt = (PremuliplyDataContext*)context;
    Pixel_8 * src = (Pixel_8 *)cnt->source->data + y * cnt->source->rowBytes;
    Pixel_8 * dst = (Pixel_8 *)cnt->dest->data + y * cnt->dest->rowBytes;
    vImagePixelCount w = cnt->source->width;

    for (int x = 0; x < w; ++x, src += 4, dst += 4) {
        const int a = src[0];
        dst[0] = a;
        dst[1] = saturate_cast_to_Pixel_8( (src[1] * a + 127) / 255 );
        dst[2] = saturate_cast_to_Pixel_8( (src[2] * a + 127) / 255 );
        dst[3] = saturate_cast_to_Pixel_8( (src[3] * a + 127) / 255 );
    }
}

vImage_Error vImagePremultiplyData_ARGB8888( const vImage_Buffer *source,
                                             const vImage_Buffer *dest,
                                             vImage_Flags flags ) {
    vImage_Error error = vImageValidateBuffer(dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    PremuliplyDataContext context;
    context.source = source;
    context.dest = dest;

    dispatch_parallel(parallel_vImagePremultiplyData_ARGB8888, (int)h, &context);

    return kvImageNoError;
}

static void parallel_vImagePremultiplyData_RGBA8888(void* context, size_t iter) {
    const size_t y = iter;
    const PremuliplyDataContext* cnt = (PremuliplyDataContext*)context;
    Pixel_8 * src = (Pixel_8 *)cnt->source->data + y * cnt->source->rowBytes;
    Pixel_8 * dst = (Pixel_8 *)cnt->dest->data + y * cnt->dest->rowBytes;
    vImagePixelCount w = cnt->source->width;

    for (int x = 0; x < w; ++x, src += 4, dst += 4) {
        const int a = src[3];
        dst[3] = a;
        dst[0] = saturate_cast_to_Pixel_8( (src[0] * a + 127) / 255 );
        dst[1] = saturate_cast_to_Pixel_8( (src[1] * a + 127) / 255 );
        dst[2] = saturate_cast_to_Pixel_8( (src[2] * a + 127) / 255 );
    }
}

vImage_Error vImagePremultiplyData_RGBA8888( const vImage_Buffer *source,
                                             const vImage_Buffer *dest,
                                             vImage_Flags flags ) {
    vImage_Error error = vImageValidateBuffer(dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    PremuliplyDataContext context;
    context.source = source;
    context.dest = dest;

#if EFFECTS_USE_GCD
    dispatch_apply_f(h, dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), &context, parallel_vImagePremultiplyData_RGBA8888);
#else
#pragma omp parallel for
    for (int iter = 0; iter < h; ++iter)
        parallel_vImagePremultiplyData_RGBA8888(&context, iter);
#endif

    return kvImageNoError;
}

static void parallel_vImageUnpremultiplyData_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const PremuliplyDataContext* cnt = (PremuliplyDataContext*)context;
    Pixel_8 * src = (Pixel_8 *)cnt->source->data + y * cnt->source->rowBytes;
    Pixel_8 * dst = (Pixel_8 *)cnt->dest->data + y * cnt->dest->rowBytes;
    vImagePixelCount w = cnt->source->width;

    for (int x = 0; x < w; ++x, src += 4, dst += 4) {
        const int a = src[0];
        dst[0] = a;
        if (a > 0) {
            const int a2 = a + a;
            dst[1] = saturate_cast_to_Pixel_8( (src[1] * 510 + a) / a2 );
            dst[2] = saturate_cast_to_Pixel_8( (src[2] * 510 + a) / a2 );
            dst[3] = saturate_cast_to_Pixel_8( (src[3] * 510 + a) / a2 );
        }
        else {
            dst[1] = src[1];
            dst[2] = src[2];
            dst[3] = src[3];
        }
    }
}

vImage_Error vImageUnpremultiplyData_ARGB8888( const vImage_Buffer *source,
                                               const vImage_Buffer *dest,
                                               vImage_Flags flags ) {
    vImage_Error error = vImageValidateBuffer(dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    PremuliplyDataContext context;
    context.source = source;
    context.dest = dest;

#if EFFECTS_USE_GCD
    dispatch_apply_f(h, dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), &context, parallel_vImageUnpremultiplyData_ARGB8888);
#else
#pragma omp parallel for
    for (int iter = 0; iter < h; ++iter)
        parallel_vImageUnpremultiplyData_ARGB8888(&context, iter);
#endif

    return kvImageNoError;
}

static void parallel_vImageUnpremultiplyData_RGBA8888(void* context, size_t iter) {
    const size_t y = iter;
    const PremuliplyDataContext* cnt = (PremuliplyDataContext*)context;
    Pixel_8 * src = (Pixel_8 *)cnt->source->data + y * cnt->source->rowBytes;
    Pixel_8 * dst = (Pixel_8 *)cnt->dest->data + y * cnt->dest->rowBytes;
    vImagePixelCount w = cnt->source->width;

    for (int x = 0; x < w; ++x, src += 4, dst += 4) {
        const int a = src[3];
        dst[3] = a;
        if (a > 0) {
            const int a2 = a + a;
            dst[0] = saturate_cast_to_Pixel_8( (src[0] * 510 + a) / a2 );
            dst[1] = saturate_cast_to_Pixel_8( (src[1] * 510 + a) / a2 );
            dst[2] = saturate_cast_to_Pixel_8( (src[2] * 510 + a) / a2 );
        }
        else {
            dst[0] = src[0];
            dst[1] = src[1];
            dst[2] = src[2];
        }
    }
}

vImage_Error vImageUnpremultiplyData_RGBA8888( const vImage_Buffer *source,
                                               const vImage_Buffer *dest,
                                               vImage_Flags flags ) {
    vImage_Error error = vImageValidateBuffer(dest);
    if (error) return error;

    const vImagePixelCount h = dest->height;

    PremuliplyDataContext context;
    context.source = source;
    context.dest = dest;

#if EFFECTS_USE_GCD
    dispatch_apply_f(h, dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), &context, parallel_vImageUnpremultiplyData_RGBA8888);
#else
#pragma omp parallel for
    for (int iter = 0; iter < h; ++iter)
        parallel_vImageUnpremultiplyData_RGBA8888(&context, iter);
#endif

    return kvImageNoError;
}


typedef struct SelectChannelsContext_ {
    const vImage_Buffer *newSrc;
    const vImage_Buffer *dest;
    uint8_t copyMask;
} SelectChannelsContext;

static void parallel_vImageSelectChannels_ARGB8888(void* context, size_t iter) {
    const size_t y = iter;
    const SelectChannelsContext* cnt = (SelectChannelsContext*)context;
    vImagePixelCount rowBytes = cnt->newSrc->rowBytes;
    Pixel_8 *newSrc = ((Pixel_8 *)cnt->newSrc->data) + rowBytes * y;

    Pixel_8 *dst = ((Pixel_8 *)cnt->dest->data) + rowBytes * y;
    uint8_t copyMask = cnt->copyMask;

    vImagePixelCount w = cnt->newSrc->width;
    for (int x = 0; x < w; ++x, newSrc += 4, dst += 4) {

        switch (copyMask) {
            case 0x1:
                dst[3] = newSrc[3];
                break;
            case 0x2:
                dst[2] = newSrc[2];
                break;
            case 0x4:
                dst[1] = newSrc[1];
                break;
            case 0x8:
                dst[0] = newSrc[0];
                break;

        }
    }
}


/*vImage_Error vImageSelectChannels_ARGB8888(
    const vImage_Buffer *newSrc,
    const vImage_Buffer *origSrc,
    const vImage_Buffer *dest,
    uint8_t copyMask,
    vImage_Flags flags)
{
   vImage_Error error = vImageValidateBuffer(dest);
   if (error) return error;
    image_copy(origSrc, dest);
    SelectChannelsContext context;

    context.newSrc = newSrc;
    context.dest = dest;
    context.copyMask = copyMask;

    const vImagePixelCount h = dest->height;

    dispatch_parallel(parallel_vImageSelectChannels_ARGB8888, (int)h, &context);

    return kvImageNoError;
}*/
