#ifndef ACCELERATE_IMAGE_H
#define ACCELERATE_IMAGE_H

#include <string.h>
#include <stdint.h>
#include <sys/types.h>


#ifdef __cplusplus
extern "C" {
#endif


#ifdef __ANDROID__
typedef unsigned short ushort;
typedef unsigned char uchar;
#endif // Android

#ifdef _WIN32
#define M_PI        3.14159265358979323846264338327950288   /* pi             */
#define M_PI_2      1.57079632679489661923132169163975144   /* pi/2           */
#endif // Windows

enum
{
   kvImageNoFlags                  =   0,
   kvImageLeaveAlphaUnchanged      =   1,
   kvImageCopyInPlace              =   2,
   kvImageBackgroundColorFill      =   4,
   kvImageEdgeExtend               =   8,
   kvImageDoNotTile                =   16,
   kvImageHighQualityResampling    =   32,
   kvImageTruncateKernel           =   64,
   kvImageGetTempBufferSize        =   128
};
typedef uint32_t vImage_Flags;


enum
{
   kvImageNoError                  =   0,
   kvImageRoiLargerThanInputBuffer =   -21766,
   kvImageInvalidKernelSize        =   -21767,
   kvImageInvalidEdgeStyle         =   -21768,
   kvImageInvalidOffset_X          =   -21769,
   kvImageInvalidOffset_Y          =   -21770,
   kvImageMemoryAllocationError    =   -21771,
   kvImageNullPointerArgument      =   -21772,
   kvImageInvalidParameter         =   -21773,
   kvImageBufferSizeMismatch       =   -21774,
   kvImageUnknownFlagsBit           =    -21775
};
#ifdef _WIN32
typedef long long vImage_Error;
#else
typedef ssize_t vImage_Error;
#endif


typedef unsigned long vImagePixelCount;       /* Pedantic: A number of pixels. For LP64 (ppc64) this is a 64-bit quantity.  */



typedef struct vImage_Buffer {
    /// Pointer to the top left pixel of the buffer.
    void *data;

    /// The height (in pixels) of the buffer.
    vImagePixelCount height;

    /// The width (in pixels) of the buffer.
    vImagePixelCount width;

    /// The number of bytes in a pixel row.
    size_t rowBytes;
} vImage_Buffer;




typedef struct vImage_Rect {
    vImagePixelCount x;
    vImagePixelCount y;
    vImagePixelCount w;
    vImagePixelCount h;
} vImage_Rect;



/// Size of CVLImageBuffer.
#define vImage_Buffer_sz sizeof(vImage_Buffer)



/// 8 bit planar pixel value.
typedef uint8_t Pixel_8;

/// Floating point planar pixel value.
typedef float Pixel_F;

/// Floating point (double precision) planar pixel value.
typedef double Pixel_D;

/// Interleaved 4 channels (8 bit/channel) pixel value.
typedef uint8_t Pixel_8888[4];

/// Interleaved 4 channels (32 bit/channel, floating point) pixel value.
typedef float Pixel_FFFF[4];

/// Interleaved 4 channels (64 bit/channel, floating point) pixel value.
typedef double Pixel_DDDD[4];



/// Size of Pixel_8.
#define Pixel_8_sz sizeof(Pixel_8)

/// Size of Pixel_F.
#define Pixel_F_sz sizeof(Pixel_F)

/// Size of Pixel_D.
#define Pixel_D_sz sizeof(Pixel_D)

/// Size of Pixel_8888.
#define Pixel_8888_sz sizeof(Pixel_8888)

/// Size of Pixel_FFFF.
#define Pixel_FFFF_sz sizeof(Pixel_FFFF)

/// Size of Pixel_DDDD.
#define Pixel_DDDD_sz sizeof(Pixel_DDDD)



#define GET_LINE_8(image, y) \
((Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_LINE_8888(image, y) \
(Pixel_8888*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_LINE_F(image, y) \
(Pixel_F*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_LINE_FFFF(image, y) \
(Pixel_FFFF*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_LINE_D(image, y) \
(Pixel_D*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_LINE_DDDD(image, y) \
(Pixel_DDDD*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)



#define GET_CONST_LINE_8(image, y) \
((const Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_CONST_LINE_8888(image, y) \
(const Pixel_8888*)((const Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_CONST_LINE_F(image, y) \
(const Pixel_F*)((const Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_CONST_LINE_FFFF(image, y) \
(const Pixel_FFFF*)((const Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_CONST_LINE_D(image, y) \
(Pixel_D*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)

#define GET_CONST_LINE_DDDD(image, y) \
(Pixel_DDDD*)((Pixel_8*)(image)->data + (y) * (image)->rowBytes)



#ifdef _WIN32
#define GET_SUB_IMAGE(image, rect, line_macros) \
{line_macros((image), (rect).y) + (rect).x, (rect).h, (rect).w, (image)->rowBytes}
#else
#define GET_SUB_IMAGE(image, rect, line_macros) \
(vImage_Buffer){line_macros((image), (rect).y) + (rect).x, (rect).h, (rect).w, (image)->rowBytes}
#endif


#define CLAMP(val, minval, maxval) \
(val) < (minval) ? (minval) : ((val) > (maxval) ? (maxval) : (val))



// Blending

/**
 * Performs premultiplied alpha compositing of two ARGB8888 images, using a single alpha value for the whole image and placing the result in a destination buffer.
 * @param srcTop pointer to a vImage buffer structure that contains data for the top source image.
 * @param constAlpha alpha value you want to apply to the image.
 * @param srcBottom pointer to a vImage buffer structure that contains data for the bottom source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes described in vImage Data Types and Constants Reference.
 */
vImage_Error vImagePremultipliedConstAlphaBlend_ARGB8888(
    const vImage_Buffer *srcTop,
    Pixel_8 constAlpha,
    const vImage_Buffer *srcBottom,
    const vImage_Buffer *dest,
    vImage_Flags flags);

/**
 * Performs nonpremultiplied alpha compositing of two ARGB8888 images, placing the result in a destination buffer.
 * @param srcTop pointer to a vImage buffer structure that contains data for the top source image.
 * @param srcBottom pointer to a vImage buffer structure that contains data for the bottom source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes described in vImage Data Types and Constants Reference.
 */
vImage_Error vImageAlphaBlend_ARGB8888 (
   const vImage_Buffer *srcTop,
   const vImage_Buffer *srcBottom,
   const vImage_Buffer *dest,
   vImage_Flags flags);

/**
 * Performs mixed alpha compositing of a nonpremultiplied Planar8 image over a premultiplied Planar8 image, placing the premultiplied result in a destination buffer.
 * @param srcTop pointer to a vImage buffer structure that contains data for the top source image.
 * @param srcTopAlpha pointer to a vImage buffer structure that contains data for the alpha values of the top source image.
 * @param srcBottom pointer to a vImage buffer structure that contains data for the bottom source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes described in vImage Data Types and Constants Reference.
 */
vImage_Error vImageAlphaBlend_NonpremultipliedToPremultiplied_Planar8 (
   const vImage_Buffer *srcTop,
   const vImage_Buffer *srcTopAlpha,
   const vImage_Buffer *srcBottom,
   const vImage_Buffer *dest,
   vImage_Flags flags
);



// Overwrite channels

/**
 * Overwrites a Planar8 image buffer with the provided value.
 * @param scalar 8-bit pixel value.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageOverwriteChannelsWithScalar_Planar8(
    Pixel_8 scalar,
    const vImage_Buffer *dest,
    vImage_Flags flags);

/**
 * Overwrites a PlanarF image buffer with the provided value.
 * @param scalar floating-point pixel value.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageOverwriteChannelsWithScalar_PlanarF(
    Pixel_F scalar,
    const vImage_Buffer *dest,
    vImage_Flags flags);

/**
 * Overwrites an ARGB8888 image buffer with the provided pixel value.
 * @param the_pixel ARGB pixel value.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to overwrite.
 * @param dest pointer to a vImage buffer data structure.
 * @param copyMask output value that selects the plane (or planes) from the ARGB8888 source buffer that you want replaced with the pixel value.
 * The value 0x8 selects the alpha channel, 0x4 the red channel, 0x2 the green channel, and 0x1 the blue channel. You can add these values together to select multiple channels.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageOverwriteChannelsWithPixel_ARGB8888(
    const Pixel_8888 the_pixel,
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    uint8_t copyMask,
    vImage_Flags flags);

/**
 * Overwrites one or more planes of an ARGB8888 image buffer with the provided planar buffer.
 * @param newSrc pointer to a vImage buffer structure that contains the data, in Planar8 format, for overwriting the origSrc image data.
 * @param origSrc pointer to a vImage buffer structure that contains the source image whose data you want to overwrite.
 * @param dest pointer to a vImage buffer data structure.
 * @param copyMask output value that selects the plane (or planes) from the ARGB8888 source buffer that you want replaced with data from the Planar8 source buffer.
 * The value 0x8 selects the alpha channel, 0x4 the red channel, 0x2 the green channel, and 0x1 the blue channel. You can add these values together to select multiple channels.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageOverwriteChannels_ARGB8888(
    const vImage_Buffer *newSrc,
    const vImage_Buffer *origSrc,
    const vImage_Buffer *dest,
    uint8_t copyMask,
    vImage_Flags flags);

/**
 * Reorders the channels in an ARGB8888 image.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to permute.
 * @param dest pointer to a vImage buffer data structure.
 * @param permuteMap array of four 8-bit integers with the values 0, 1, 2, and 3, in some order.
 * Each value specifies a plane from the source image that should be copied to that plane in the destination image.
 * 0 denotes the alpha channel, 1 the red channel, 2 the green channel, and 3 the blue channel.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImagePermuteChannels_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const uint8_t permuteMap[4],
    vImage_Flags flags);



// Lookup table

/**
 * Transforms an Planar8 image by substituting pixel values with pixel values provided by four lookup tables.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to transform.
 * @param dest pointer to a vImage buffer data structure.
 * @param table lookup table to use.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageTableLookUp_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const Pixel_8 table[256],
    vImage_Flags flags);

/**
 * Transforms an ARGB8888 image by substituting pixel values with pixel values provided by four lookup tables.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to transform.
 * @param dest pointer to a vImage buffer data structure.
 * @param alphaTable lookup table to use for the alpha channel of the source image. If you pass NULL for this table, the function copies the alpha channel unchanged to the destination buffer.
 * @param redTable lookup table to use for the red channel of the source image. If you pass NULL for this table, the function copies the alpha channel unchanged to the destination buffer.
 * @param greenTable lookup table to use for the green channel of the source image. If you pass NULL for this table, the function copies the alpha channel unchanged to the destination buffer.
 * @param blueTable lookup table to use for the blue channel of the source image. If you pass NULL for this table, the function copies the alpha channel unchanged to the destination buffer.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageTableLookUp_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const Pixel_8 alphaTable[256],
    const Pixel_8 redTable[256],
    const Pixel_8 greenTable[256],
    const Pixel_8 blueTable[256],
    vImage_Flags flags);



// Scale image

/**
 * Scales an ARGB8888 source image to fit a destination buffer.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to scale.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageScale_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImage_Flags flags);

/**
 * Scales a Planar8 source image to fit a destination buffer.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to scale.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageScale_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImage_Flags flags);

/**
 * Scales a PlanarF source image to fit a destination buffer.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to scale.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageScale_PlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImage_Flags flags);


// Histogram calculation

/**
 * Calculates a histogram for a Planar8 image.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param histogram pointer to a histogram. On return, this array will contain the histogram for the source image. The histogram will be an array with 256 elements.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageHistogramCalculation_Planar8(
    const vImage_Buffer *src,
    vImagePixelCount *histogram,
    vImage_Flags flags);

/**
 * Calculates a histogram for an ARGB8888 image.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param histogram pointer to a histograms, one each for alpha, red, green, and blue (in that order).
 * On return, this array will contain the four histograms for the corresponding channels.
 * Each of the four histograms will be an array with 256 elements.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageHistogramCalculation_ARGB8888(
    const vImage_Buffer *src,
    vImagePixelCount *histogram[4],
    vImage_Flags flags);

/**
 * Stretches the contrast of a Planar8 source image.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to scale.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageContrastStretch_Planar8 (
   const vImage_Buffer *src,
   const vImage_Buffer *dest,
   vImage_Flags flags);



// Format conversion

/**
 * Converts a four-channel ARGB8888 image to a three-channel RGB888 image by removing the alpha channel.
 * @param argbSrc pointer to a vImage buffer structure that contains the source image whose data you want to convert.
 * @param rgbDest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageConvert_ARGB8888toRGB888(
    const vImage_Buffer *argbSrc,
    const vImage_Buffer *rgbDest,
    vImage_Flags flags);

/**
 * Interleaves three Planar8 images into one XRGB8888 image.
 * @param alpha value to use for the alpha channel of the destination image.
 * @param red pointer to a vImage buffer structure that contains the Planar8 image to use as the red channel of the destination image.
 * @param green pointer to a vImage buffer structure that contains the Planar8 image to use as the green channel of the destination image.
 * @param blue pointer to a vImage buffer structure that contains the Planar8 image to use as the blue channel of the destination image.
 * @param dest pointer to a vImage buffer data structure.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageConvert_Planar8ToXRGB8888(
    Pixel_8 alpha,
    const vImage_Buffer *red,
    const vImage_Buffer *green,
    const vImage_Buffer *blue,
    const vImage_Buffer *dest,
    vImage_Flags flags);

/**
 * Converts a Planar8 image to a PlanarF image.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to convert.
 * @param dest pointer to a vImage buffer data structure.
 * @param maxFloat maximum pixel value for the destination image.
 * @param minFloat minimum pixel value for the destination image.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageConvert_Planar8toPlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    Pixel_F maxFloat,
    Pixel_F minFloat,
    vImage_Flags flags);



// Convolve

/**
 * Convolves a region of interest within a Planar8 source image by an implicit M x N kernel that has the effect of a box filter.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param srcOffsetToROI_X the horizontal offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param srcOffsetToROI_Y the vertical offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param kernel_height the height of the kernel in pixels. This value must be odd.
 * @param kernel_width the width of the kernel in pixels. This value must be odd.
 * @param backgroundColor a background color. Not used.
 * @param flags not used. For now functions works such as flags == kvImageTruncateKernel.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageBoxConvolve_Planar8(
   const vImage_Buffer *src,
   const vImage_Buffer *dest,
   void *tempBuffer,
   vImagePixelCount srcOffsetToROI_X,
   vImagePixelCount srcOffsetToROI_Y,
   uint32_t kernel_height,
   uint32_t kernel_width,
   Pixel_8 backgroundColor,
   vImage_Flags flags);

/**
 * Convolves a region of interest within an ARGB8888 source image by an implicit M x N kernel that has the effect of a box filter.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param srcOffsetToROI_X the horizontal offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param srcOffsetToROI_Y the vertical offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param kernel_height the height of the kernel in pixels. This value must be odd.
 * @param kernel_width the width of the kernel in pixels. This value must be odd.
 * @param backgroundColor a background color. Not used.
 * @param flags not used. For now functions works such as flags == kvImageTruncateKernel | kvImageLeaveAlphaUnchanged.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageBoxConvolve_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_8888 backgroundColor,
    vImage_Flags flags);

/**
 * Convolves a region of interest within a Planar8 source image by an implicit M x N kernel that has the effect of a tent filter.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param srcOffsetToROI_X the horizontal offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param srcOffsetToROI_Y the vertical offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param kernel_height the height of the kernel in pixels. This value must be odd.
 * @param kernel_width the width of the kernel in pixels. This value must be odd.
 * @param backgroundColor a background color. Not used.
 * @param flags not used. For now functions works such as flags == kvImageTruncateKernel.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageTentConvolve_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_8 backgroundColor,
    vImage_Flags flags);

/**
 * Convolves a region of interest within an ARGB8888 source image by an implicit M x N kernel that has the effect of a tent filter.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param srcOffsetToROI_X the horizontal offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param srcOffsetToROI_Y the vertical offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param kernel_height the height of the kernel in pixels. This value must be odd.
 * @param kernel_width the width of the kernel in pixels. This value must be odd.
 * @param backgroundColor a background color. Not used.
 * @param flags not used. For now functions works such as flags == kvImageTruncateKernel | kvImageLeaveAlphaUnchanged.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageTentConvolve_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_8888 backgroundColor,
    vImage_Flags flags);

/**
 * Convolves a region of interest within a source image by an M x N kernel.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer pointer to a temporary buffer. Not used.
 * @param srcOffsetToROI_X the horizontal offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param srcOffsetToROI_Y the vertical offset, in pixels, to the upper-left pixel of the region of interest within the source image.
 * @param kernel pointer to the convolution kernel data, which must be a packed array without any padding.
 * @param kernel_height the height of the kernel in pixels. This value must be odd.
 * @param kernel_width the width of the kernel in pixels. This value must be odd.
 * @param backgroundColor a background color. Not used.
 * @param flags not used. For now functions works such as flags == kvImageEdgeExtend.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageConvolve_PlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    vImagePixelCount srcOffsetToROI_X,
    vImagePixelCount srcOffsetToROI_Y,
    const float *kernel,
    uint32_t kernel_height,
    uint32_t kernel_width,
    Pixel_F backgroundColor,
    vImage_Flags flags);



// Matrix multiply

/**
 * Operates upon an interleaved 8-bit source image, multiplying each pixel by the provided matrix to produce an interleaved 8-bit destination image.
 * @param src pointer to a vImage buffer structure that contains the source image.
 * @param dest pointer to a vImage buffer data structure.
 * @param matrix 1-dimensional array whose values represent a 4x4 matrix. vImage multiplies each source pixel by this matrix to produce a destination pixel.
 * @param divisor divisor for normalization after performing the matrix multiplication.
 * @param pre_bias packed array of bias values, one for each source plane. vImage adds the appropriate bias value to each source value prior to matrix multiplication.
 * Pass NULL if you do not want to apply a preprocessing bias value.
 * @param post_bias packed array of bias values, one for each destination plane. vImage adds the appropriate bias value to each destination value after matrix multiplication.
 * Pass NULL if you do not want to apply a preprocessing bias value.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageMatrixMultiply_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    const int16_t matrix[16],
    int32_t divisor,
    const int16_t *pre_bias,
    const int32_t *post_bias,
    vImage_Flags flags);



// Clip

/**
 * Clips the pixel values of an image in PlanarF format, using the provided minimum and maximum values.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to clip.
 * @param dest pointer to a vImage buffer data structure.
 * @param maxFloat maximum pixel value. The function clips larger values to this value in the destination image.
 * @param minFloat minimum pixel value. The function clips smaller values to this value in the destination image.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageClip_PlanarF(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    Pixel_F maxFloat,
    Pixel_F minFloat,
    vImage_Flags flags);



// Rotate

/**
 * Rotates an Planar8 source image by the provided factor of 90.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to rotate.
 * @param dest pointer to a vImage buffer data structure.
 * @param rotationConstant value specifying the angle of rotation.
 * @param backColor background color. Not used.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageRotate90_Planar8(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    uint8_t rotationConstant,
    Pixel_8 backColor,
    vImage_Flags flags);

/**
 * Rotates an ARGB8888 source image by the provided factor of 90.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to rotate.
 * @param dest pointer to a vImage buffer data structure.
 * @param rotationConstant value specifying the angle of rotation.
 * @param backColor background color. Not used.
 * @param flags not used.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageRotate90_ARGB8888(
    const vImage_Buffer *src,
    const vImage_Buffer *dest,
    uint8_t rotationConstant,
    Pixel_8888 backColor,
    vImage_Flags flags);

/**
 * Rotates a Planar8 source image by the provided angle.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to rotate.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer not used.
 * @param angleInRadians angle of rotation, in radians.
 * @param backColor background color. Pass a pixel value only if you also set the kvImageBackgroundColorFill flag.
 * @param flags kvImageBackgroundColorFill or kvImageEdgeExtend.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageRotate_Planar8 (
    const vImage_Buffer  *src,
    const vImage_Buffer *dest,
    void *tempBuffer,
    float angleInRadians,
    Pixel_8 backColor,
    vImage_Flags flags);

/**
 * Rotates an ARGB8888 source image by the provided angle.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to rotate.
 * @param dest pointer to a vImage buffer data structure.
 * @param tempBuffer not used.
 * @param angleInRadians angle of rotation, in radians.
 * @param backColor background color. Pass a pixel value only if you also set the kvImageBackgroundColorFill flag.
 * @param flags kvImageBackgroundColorFill or kvImageEdgeExtend.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageRotate_ARGB8888 (
    const vImage_Buffer* src,
    vImage_Buffer* dest,
    void *tempBuffer,
    float angleInRadians,
    Pixel_8888 backColor,
    vImage_Flags flags);


/**
 * Reflects an ARGB8888 source image left to right across the center vertical line of the image.
 * @param src pointer to a vImage buffer structure that contains the source image whose data you want to reflect..
 * @param dest pointer to a vImage buffer data structure.
 * @param flags kvImageNoFlags.
 * @return kvImageNoError, otherwise one of the error codes.
 */
vImage_Error vImageHorizontalReflect_ARGB8888 (
    const vImage_Buffer* source,
    vImage_Buffer* dest,
    vImage_Flags flags);

vImage_Error vImagePremultiplyData_ARGB8888(
    const vImage_Buffer *source,
    const vImage_Buffer *dest,
    vImage_Flags flags );

vImage_Error vImagePremultiplyData_RGBA8888(
    const vImage_Buffer *source,
    const vImage_Buffer *dest,
    vImage_Flags flags );

vImage_Error vImageUnpremultiplyData_ARGB8888(
    const vImage_Buffer *source,
    const vImage_Buffer *dest,
    vImage_Flags flags );

vImage_Error vImageUnpremultiplyData_RGBA8888(
    const vImage_Buffer *source,
    const vImage_Buffer *dest,
    vImage_Flags flags );

vImage_Error vImageSelectChannels_ARGB8888(
    const vImage_Buffer *newSrc,       /* A ARGB interleaved buffer */
    const vImage_Buffer *origSrc,      /* A ARGB interleaved buffer */
    const vImage_Buffer *dest,      /* A ARGB interleaved buffer */
    uint8_t copyMask,               /* Copy plane into  0x8  -- alpha, 0x4 -- red, 0x2 --- green, 0x1 --- blue */
    vImage_Flags    flags );

#ifdef __cplusplus
}
#endif


#endif //ACCELERATE_IMAGE_H
