#include "resize.h"

#include <algorithm>
#include <cmath>
#include <float.h>
#include <limits.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define PI   3.1415926535897932384626433832795

static const int MAX_ESIZE=16;

#ifdef _WIN32
    static double round(double d) {
        return floor(d + 0.5);
    }
    static float round(float d) {
        return floor(d + 0.5f);
    }
#endif

struct VResizeNoVec
{
    int operator()(const unsigned char**, unsigned char*, const unsigned char*, int) const { return 0; }
};

struct HResizeNoVec
{
    int operator()(const unsigned char**, unsigned char**, int, const int*,
        const unsigned char*, int, int, int, int, int) const { return 0; }
};

typedef VResizeNoVec VResizeCubicVec_32s8u;
typedef VResizeNoVec VResizeCubicVec_32f16u;
typedef VResizeNoVec VResizeCubicVec_32f16s;
typedef VResizeNoVec VResizeCubicVec_32f;


template<typename _Tp> static inline _Tp saturate_cast(unsigned char v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(char v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(unsigned short v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(short v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(unsigned v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(int v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(float v) { return _Tp(v); }
template<typename _Tp> static inline _Tp saturate_cast(double v) { return _Tp(v); }

template<> inline unsigned char saturate_cast<unsigned char>(char v)
{ return (unsigned char)std::max((int)v, 0); }
template<> inline unsigned char saturate_cast<unsigned char>(unsigned short v)
{ return (unsigned char)std::min((unsigned)v, (unsigned)UCHAR_MAX); }
template<> inline unsigned char saturate_cast<unsigned char>(int v)
{ return (unsigned char)((unsigned)v <= UCHAR_MAX ? v : v > 0 ? UCHAR_MAX : 0); }
template<> inline unsigned char saturate_cast<unsigned char>(short v)
{ return saturate_cast<unsigned char>((int)v); }
template<> inline unsigned char saturate_cast<unsigned char>(unsigned v)
{ return (unsigned char)std::min(v, (unsigned)UCHAR_MAX); }
template<> inline unsigned char saturate_cast<unsigned char>(float v)
{ int iv = round(v); return saturate_cast<unsigned char>(iv); }
template<> inline unsigned char saturate_cast<unsigned char>(double v)
{ int iv = round(v); return saturate_cast<unsigned char>(iv); }

template<> inline char saturate_cast<char>(unsigned char v)
{ return (char)std::min((int)v, SCHAR_MAX); }
template<> inline char saturate_cast<char>(unsigned short v)
{ return (char)std::min((unsigned)v, (unsigned)SCHAR_MAX); }
template<> inline char saturate_cast<char>(int v)
{
    return (char)((unsigned)(v-SCHAR_MIN) <= (unsigned)UCHAR_MAX ?
                v : v > 0 ? SCHAR_MAX : SCHAR_MIN);
}
template<> inline char saturate_cast<char>(short v)
{ return saturate_cast<char>((int)v); }
template<> inline char saturate_cast<char>(unsigned v)
{ return (char)std::min(v, (unsigned)SCHAR_MAX); }

template<> inline char saturate_cast<char>(float v)
{ int iv = round(v); return saturate_cast<char>(iv); }
template<> inline char saturate_cast<char>(double v)
{ int iv = round(v); return saturate_cast<char>(iv); }

template<> inline unsigned short saturate_cast<unsigned short>(char v)
{ return (unsigned short)std::max((int)v, 0); }
template<> inline unsigned short saturate_cast<unsigned short>(short v)
{ return (unsigned short)std::max((int)v, 0); }
template<> inline unsigned short saturate_cast<unsigned short>(int v)
{ return (unsigned short)((unsigned)v <= (unsigned)USHRT_MAX ? v : v > 0 ? USHRT_MAX : 0); }
template<> inline unsigned short saturate_cast<unsigned short>(unsigned v)
{ return (unsigned short)std::min(v, (unsigned)USHRT_MAX); }
template<> inline unsigned short saturate_cast<unsigned short>(float v)
{ int iv = round(v); return saturate_cast<unsigned short>(iv); }
template<> inline unsigned short saturate_cast<unsigned short>(double v)
{ int iv = round(v); return saturate_cast<unsigned short>(iv); }

template<> inline short saturate_cast<short>(unsigned short v)
{ return (short)std::min((int)v, SHRT_MAX); }
template<> inline short saturate_cast<short>(int v)
{
    return (short)((unsigned)(v - SHRT_MIN) <= (unsigned)USHRT_MAX ?
            v : v > 0 ? SHRT_MAX : SHRT_MIN);
}
template<> inline short saturate_cast<short>(unsigned v)
{ return (short)std::min(v, (unsigned)SHRT_MAX); }
template<> inline short saturate_cast<short>(float v)
{ int iv = round(v); return saturate_cast<short>(iv); }
template<> inline short saturate_cast<short>(double v)
{ int iv = round(v); return saturate_cast<short>(iv); }

template<> inline int saturate_cast<int>(float v) { return round(v); }
template<> inline int saturate_cast<int>(double v) { return round(v); }

// we intentionally do not clip negative numbers, to make -1 become 0xffffffff etc.
template<> inline unsigned saturate_cast<unsigned>(float v){ return round(v); }
template<> inline unsigned saturate_cast<unsigned>(double v) { return round(v); }

/* ========================== Auto buffer ========================== */

template<typename _Tp, size_t fixed_size=4096/sizeof(_Tp)+8> class AutoBuffer
{
public:
    typedef _Tp value_type;
    enum { buffer_padding = (int)((16 + sizeof(_Tp) - 1)/sizeof(_Tp)) };

    //! the default contructor
    AutoBuffer();
    //! constructor taking the real buffer size
    AutoBuffer(size_t _size);
    //! destructor. calls deallocate()
    ~AutoBuffer();

    //! allocates the new buffer of size _size. if the _size is small enough, stack-allocated buffer is used
    void allocate(size_t _size);
    //! deallocates the buffer if it was dynamically allocated
    void deallocate();
    //! returns pointer to the real buffer, stack-allocated or head-allocated
    operator _Tp* ();
    //! returns read-only pointer to the real buffer, stack-allocated or head-allocated
    operator const _Tp* () const;

protected:
    //! pointer to the real buffer, can point to buf if the buffer is small enough
    _Tp* ptr;
    //! size of the real buffer
    size_t size;
    //! pre-allocated buffer
    _Tp buf[fixed_size+buffer_padding];
};


template<typename _Tp, size_t fixed_size> inline AutoBuffer<_Tp, fixed_size>::AutoBuffer() {
    ptr = buf;
    size = fixed_size;
}

template<typename _Tp, size_t fixed_size> inline AutoBuffer<_Tp, fixed_size>::AutoBuffer(size_t _size) {
    ptr = buf;
    size = fixed_size;
    allocate(_size);
}

template<typename _Tp, size_t fixed_size> inline AutoBuffer<_Tp, fixed_size>::~AutoBuffer()
{ deallocate(); }

template<typename _Tp, size_t fixed_size> inline void AutoBuffer<_Tp, fixed_size>::allocate(size_t _size) {
    if(_size <= size)
        return;
    deallocate();
    if(_size > fixed_size) {
        ptr = new _Tp[_size];
        size = _size;
    }
}

template<typename _Tp, size_t fixed_size> inline void AutoBuffer<_Tp, fixed_size>::deallocate() {
    if (ptr != buf) {
        delete [] ptr;
        ptr = buf;
        size = fixed_size;
    }
}

template<typename _Tp, size_t fixed_size> inline AutoBuffer<_Tp, fixed_size>::operator _Tp* ()
{ return ptr; }

template<typename _Tp, size_t fixed_size> inline AutoBuffer<_Tp, fixed_size>::operator const _Tp* () const
{ return ptr; }

typedef void (*ResizeFunc)(
        const void* src, int src_width, int src_height, int src_step,
        void* dst, int dst_width, int dst_height, int dst_step,
        const int cn,
        const int* xofs, const void* alpha,
        const int* yofs, const void* beta,
        int xmin, int xmax, int ksize);

template<typename ST, typename DT> struct Cast
{
    typedef ST type1;
    typedef DT rtype;

    DT operator()(ST val) const { return saturate_cast<DT>(val); }
};

template<typename ST, typename DT, int bits> struct FixedPtCast
{
    typedef ST type1;
    typedef DT rtype;
    enum { SHIFT = bits, DELTA = 1 << (bits-1) };

    DT operator()(ST val) const { return saturate_cast<DT>((val + DELTA)>>SHIFT); }
};

static inline size_t alignSize(size_t sz, int n)
{
    return (sz + n-1) & -n;
}

static inline int clip(int x, int a, int b)
{
    return x >= a ? (x < b ? x : b-1) : a;
}

struct DecimateAlpha
{
    int si, di;
    float alpha;
};

template<class HResize, class VResize>
static void resize_generic(const void* src, int src_width, int src_height, int src_step,
                                 void* dst, int dst_width, int dst_height, int dst_step,
                                 const int cn,
                                 const int* xofs, const void* _alpha,
                                 const int* yofs, const void* _beta,
                                 int xmin, int xmax, int ksize) {
    typedef typename HResize::value_type T;
    typedef typename HResize::buf_type WT;
    typedef typename HResize::alpha_type AT;

    const AT* alpha = (const AT*)_alpha;
    const AT* beta = (const AT*)_beta;
    src_width *= cn;
    dst_width *= cn;
    int bufstep = (int)alignSize(dst_width, 16); // 4*4 ?
    AutoBuffer<WT> _buffer(bufstep*ksize);
    const T* srows[MAX_ESIZE]={0};
    WT* rows[MAX_ESIZE]={0};
    int prev_sy[MAX_ESIZE];
    int k, dy;
    xmin *= cn;
    xmax *= cn;

    HResize hresize;
    VResize vresize;

    for (k = 0; k < ksize; k++) {
        prev_sy[k] = -1;
        rows[k] = (WT*)_buffer + bufstep*k;
    }

    // image resize is a separable operation. In case of not too strong
    for (dy = 0; dy < dst_height; dy++, beta += ksize) {
        int sy0 = yofs[dy], k, k0=ksize, k1=0, ksize2 = ksize/2;

        for (k = 0; k < ksize; k++) {
            int sy = clip(sy0 - ksize2 + 1 + k, 0, src_height);
            for (k1 = std::max(k1, k); k1 < ksize; k1++)
            {
                if (sy == prev_sy[k1]) {
                    if (k1 > k)
                        memcpy(rows[k], rows[k1], bufstep*sizeof(rows[0][0]));
                    break;
                }
            }
            if (k1 == ksize)
                k0 = std::min(k0, k); // remember the first row that needs to be computed
            srows[k] = (const T*)((const unsigned char*)src + src_step*sy);
            prev_sy[k] = sy;
        }

        if (k0 < ksize)
            hresize(srows + k0, rows + k0, ksize - k0, xofs, alpha,
                     src_width, dst_width, cn, xmin, xmax);

        vresize((const WT**)rows, (T*)((const unsigned char*)dst + dst_step*dy), beta, dst_width);
    }
}

const int INTER_RESIZE_COEF_BITS=11;
const int INTER_RESIZE_COEF_SCALE=1 << INTER_RESIZE_COEF_BITS;


static inline void interpolateCubic(float x, float* coeffs) {
    const float A = -0.75f;

    coeffs[0] = ((A*(x + 1) - 5*A)*(x + 1) + 8*A)*(x + 1) - 4*A;
    coeffs[1] = ((A + 2)*x - (A + 3))*x*x + 1;
    coeffs[2] = ((A + 2)*(1 - x) - (A + 3))*(1 - x)*(1 - x) + 1;
    coeffs[3] = 1.f - coeffs[0] - coeffs[1] - coeffs[2];
}

template<typename T, typename WT, typename AT>
struct HResizeCubic
{
    typedef T value_type;
    typedef WT buf_type;
    typedef AT alpha_type;

    void operator()(const T** src, WT** dst, int count,
                    const int* xofs, const AT* alpha,
                    int swidth, int dwidth, int cn, int xmin, int xmax) const
    {
        for (int k = 0; k < count; k++) {
            const T *S = src[k];
            WT *D = dst[k];
            int dx = 0, limit = xmin;
            for (;;) {
                for (; dx < limit; dx++, alpha += 4) {
                    int j, sx = xofs[dx] - cn;
                    WT v = 0;
                    for (j = 0; j < 4; j++) {
                        int sxj = sx + j*cn;
                        if ((unsigned)sxj >= (unsigned)swidth) {
                            while(sxj < 0)
                                sxj += cn;
                            while(sxj >= swidth)
                                sxj -= cn;
                        }
                        v += S[sxj]*alpha[j];
                    }
                    D[dx] = v;
                }
                if (limit == dwidth)
                    break;
                for (; dx < xmax; dx++, alpha += 4) {
                    int sx = xofs[dx];
                    D[dx] = S[sx-cn]*alpha[0] + S[sx]*alpha[1] +
                        S[sx+cn]*alpha[2] + S[sx+cn*2]*alpha[3];
                }
                limit = dwidth;
            }
            alpha -= dwidth*4;
        }
    }
};



template<typename T, typename WT, typename AT, class CastOp, class VecOp>
struct VResizeCubic
{
    typedef T value_type;
    typedef WT buf_type;
    typedef AT alpha_type;

    void operator()(const WT** src, T* dst, const AT* beta, int width) const
    {
        WT b0 = beta[0], b1 = beta[1], b2 = beta[2], b3 = beta[3];
        const WT *S0 = src[0], *S1 = src[1], *S2 = src[2], *S3 = src[3];
        CastOp castOp;
        VecOp vecOp;

        int x = vecOp((const unsigned char**)src, (unsigned char*)dst, (const unsigned char*)beta, width);
        for (; x < width; x++)
            dst[x] = castOp(S0[x]*b0 + S1[x]*b1 + S2[x]*b2 + S3[x]*b3);
    }
};

void resize_8888(const unsigned char* src, int src_width, int src_height, int src_step,
                 unsigned char* dst, int dst_width, int dst_height, int dst_step) {
    static ResizeFunc cubic = resize_generic<
            HResizeCubic<unsigned char, int, short>,
            VResizeCubic<unsigned char, int, short,
                    FixedPtCast<int, unsigned char, INTER_RESIZE_COEF_BITS * 2>,
                    VResizeCubicVec_32s8u> >;

    const int cn = 4;

    const double inv_scale_x = (double)dst_width  / src_width;
    const double inv_scale_y = (double)dst_height / src_height;
    const double scale_x = 1./inv_scale_x;
    const double scale_y = 1./inv_scale_y;

    int k, sx, sy, dx, dy;

    int xmin = 0, xmax = dst_width, width = dst_width*cn;
    float fx, fy;
    ResizeFunc func=0;
    int ksize=0, ksize2;
    ksize = 4, func = cubic;
    ksize2 = ksize/2;


    AutoBuffer<unsigned char> _buffer((width + dst_height)*(sizeof(int) + sizeof(float)*ksize));
    int* xofs = (int*)(unsigned char*)_buffer;
    int* yofs = xofs + width;
    float* alpha = (float*)(yofs + dst_height);
    short* ialpha = (short*)alpha;
    short* ibeta = ialpha + width*ksize;
    float cbuf[MAX_ESIZE];

    for (dx = 0; dx < dst_width; dx++) {

        fx = (float)((dx+0.5)*scale_x - 0.5);
        sx = floor(fx);
        fx -= sx;

        if (sx < ksize2-1) {
            xmin = dx+1;
            if (sx < 0)
                fx = 0, sx = 0;
        }

        if (sx + ksize2 >= src_width) {
            xmax = std::min(xmax, dx);
            if (sx >= src_width-1)
                fx = 0, sx = src_width-1;
        }

        for (k = 0, sx *= cn; k < cn; k++)
            xofs[dx*cn + k] = sx + k;

        interpolateCubic(fx, cbuf);

        for (k = 0; k < ksize; k++)
            ialpha[dx*cn*ksize + k] = saturate_cast<short>(cbuf[k]*INTER_RESIZE_COEF_SCALE);
        for (; k < cn*ksize; k++)
            ialpha[dx*cn*ksize + k] = ialpha[dx*cn*ksize + k - ksize];
    }

    for (dy = 0; dy < dst_height; dy++) {

        fy = (float)((dy+0.5)*scale_y - 0.5);
        sy = floor(fy);
        fy -= sy;

        yofs[dy] = sy;
        interpolateCubic(fy, cbuf);

        for (k = 0; k < ksize; k++)
            ibeta[dy*ksize + k] = saturate_cast<short>(cbuf[k]*INTER_RESIZE_COEF_SCALE);
    }

    func(src, src_width, src_height, src_step,
         dst, dst_width, dst_height, dst_step,
         4,
         xofs, (void*)ialpha, yofs,
         (void*)ibeta, xmin, xmax, ksize);
}

void resize_8(const unsigned char* src, int src_width, int src_height, int src_step,
                 unsigned char* dst, int dst_width, int dst_height, int dst_step) {
    static ResizeFunc cubic = resize_generic<
            HResizeCubic<unsigned char, int, short>,
            VResizeCubic<unsigned char, int, short,
                    FixedPtCast<int, unsigned char, INTER_RESIZE_COEF_BITS * 2>,
                    VResizeCubicVec_32s8u> >;

    const int cn = 1;

    const double inv_scale_x = (double) dst_width / src_width;
    const double inv_scale_y = (double) dst_height / src_height;
    const double scale_x = 1. / inv_scale_x;
    const double scale_y = 1. / inv_scale_y;

    int k, sx, sy, dx, dy;

    int xmin = 0, xmax = dst_width, width = dst_width * cn;
    float fx, fy;
    ResizeFunc func = 0;
    int ksize = 0, ksize2;
    ksize = 4, func = cubic;
    ksize2 = ksize / 2;

    AutoBuffer<unsigned char> _buffer((width + dst_height) * (sizeof(int) + sizeof(float) * ksize));
    int* xofs = (int*) (unsigned char*) _buffer;
    int* yofs = xofs + width;
    float* alpha = (float*) (yofs + dst_height);
    short* ialpha = (short*) alpha;
    short* ibeta = ialpha + width * ksize;
    float cbuf[MAX_ESIZE];

    for (dx = 0; dx < dst_width; dx++) {

        fx = (float) ((dx + 0.5) * scale_x - 0.5);
        sx = floor(fx);
        fx -= sx;

        if (sx < ksize2 - 1) {
            xmin = dx + 1;
            if (sx < 0) fx = 0, sx = 0;
        }

        if (sx + ksize2 >= src_width) {
            xmax = std::min(xmax, dx);
            if (sx >= src_width - 1) fx = 0, sx = src_width - 1;
        }

        for (k = 0, sx *= cn; k < cn; k++)
            xofs[dx * cn + k] = sx + k;

        interpolateCubic(fx, cbuf);

        for (k = 0; k < ksize; k++)
            ialpha[dx * cn * ksize + k] = saturate_cast<short>(cbuf[k] * INTER_RESIZE_COEF_SCALE);
        for (; k < cn * ksize; k++)
            ialpha[dx * cn * ksize + k] = ialpha[dx * cn * ksize + k - ksize];
    }

    for (dy = 0; dy < dst_height; dy++) {

        fy = (float) ((dy + 0.5) * scale_y - 0.5);
        sy = floor(fy);
        fy -= sy;

        yofs[dy] = sy;
        interpolateCubic(fy, cbuf);

        for (k = 0; k < ksize; k++)
            ibeta[dy * ksize + k] = saturate_cast<short>(cbuf[k] * INTER_RESIZE_COEF_SCALE);
    }

    func(src, src_width, src_height, src_step,
        dst, dst_width, dst_height, dst_step,
        1,
        xofs, (void*) ialpha, yofs,
        (void*) ibeta, xmin, xmax, ksize);
}

void resize_F(const float* src, int src_width, int src_height, int src_step,
                 float* dst, int dst_width, int dst_height, int dst_step) {
    static ResizeFunc cubic = resize_generic<
            HResizeCubic<float, float, float>,
            VResizeCubic<float, float, float,
                    Cast<float, float>,
                    VResizeCubicVec_32f> >;

    const int cn = 1;

    const double inv_scale_x = (double) dst_width / src_width;
    const double inv_scale_y = (double) dst_height / src_height;
    const double scale_x = 1. / inv_scale_x;
    const double scale_y = 1. / inv_scale_y;

    int k, sx, sy, dx, dy;

    int xmin = 0, xmax = dst_width, width = dst_width * cn;
    float fx, fy;
    ResizeFunc func = 0;
    int ksize = 0, ksize2;
    ksize = 4, func = cubic;
    ksize2 = ksize / 2;

    AutoBuffer<unsigned char> _buffer((width + dst_height) * (sizeof(int) + sizeof(float) * ksize));
    int* xofs = (int*) (unsigned char*) _buffer;
    int* yofs = xofs + width;
    float* alpha = (float*) (yofs + dst_height);
    float* beta = alpha + width*ksize;
    float cbuf[MAX_ESIZE];

    for (dx = 0; dx < dst_width; dx++) {

        fx = (float) ((dx + 0.5) * scale_x - 0.5);
        sx = floor(fx);
        fx -= sx;

        if (sx < ksize2 - 1) {
            xmin = dx + 1;
            if (sx < 0) fx = 0, sx = 0;
        }

        if (sx + ksize2 >= src_width) {
            xmax = std::min(xmax, dx);
            if (sx >= src_width - 1) fx = 0, sx = src_width - 1;
        }

        for (k = 0, sx *= cn; k < cn; k++)
            xofs[dx * cn + k] = sx + k;

        interpolateCubic(fx, cbuf);

        for (k = 0; k < ksize; k++)
            alpha[dx * cn * ksize + k] = cbuf[k];
        for (; k < cn * ksize; k++)
            alpha[dx * cn * ksize + k] = alpha[dx * cn * ksize + k - ksize];
    }

    for (dy = 0; dy < dst_height; dy++) {

        fy = (float) ((dy + 0.5) * scale_y - 0.5);
        sy = floor(fy);
        fy -= sy;

        yofs[dy] = sy;
        interpolateCubic(fy, cbuf);

        for (k = 0; k < ksize; k++)
            beta[dy * ksize + k] = cbuf[k];
    }

    func(src, src_width, src_height, src_step,
        dst, dst_width, dst_height, dst_step,
        1,
        xofs, (void*) alpha, yofs,
        (void*) beta, xmin, xmax, ksize);
}
