#include "warp.h"

#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#ifdef _WIN32
typedef uint16 ushort;
static long int lrint (double x) {
	return (long int)(x + 0.5);
}
#endif

static inline int min(int a, int b) {
    return b < a ? b : a;
}

static inline int max(int a, int b) {
    return b > a ? b : a;
}

static inline int saturate_cast_to_int(Pixel_D v){
    if (v < INT_MIN)
        return INT_MIN;
    if (v > INT_MAX)
        return INT_MAX;
    return (int)lrint(v);
}

static inline short saturate_cast_to_short(int v) {
    if (v <= SHRT_MIN)
        return SHRT_MIN;
    if (v <= SHRT_MAX)
        return v;
    return SHRT_MAX;
}

static inline Pixel_8 saturate_cast_to_Pixel_8(int v) {
    if ((unsigned int)v <= 255)
        return v;
    if (v < 0)
        return 0;
    return 255;
}

/////////////////====================================////////////////////////
static const int INTER_BITS     = 5;
static const int INTER_TAB_SIZE = 1 << INTER_BITS;
static const int INTER_REMAP_COEF_BITS = 15;
static const int INTER_REMAP_COEF_SCALE = 1 << INTER_REMAP_COEF_BITS;
static const int INTER_TAB_SIZE2 = INTER_TAB_SIZE * INTER_TAB_SIZE;

static unsigned char NNDeltaTab_i [INTER_TAB_SIZE2][2];
static float BilinearTab_f[INTER_TAB_SIZE2][2][2];
static short BilinearTab_i[INTER_TAB_SIZE2][2][2];
/////////////////====================================////////////////////////

static inline void interpolateLinear(Pixel_F x, Pixel_F* coeffs ) {
    coeffs[0] = 1.f - x;
    coeffs[1] = x;
}

static void initInterTab1D(Pixel_F* tab){
    Pixel_F scale = 1.f / INTER_TAB_SIZE;
    for (int i = 0; i < INTER_TAB_SIZE; i++, tab += 2) {
        interpolateLinear( i * scale, tab );
    }
}

static inline int clip(int x, int a, int b){
    return x >= a ? (x < b ? x : b-1) : a;
}

static const void* initInterTab2D() {
    static bool inittab = false;
    float*  tab = 0;
    short* itab = 0;
    int   ksize = 2;

     tab = BilinearTab_f[0][0],
    itab = BilinearTab_i[0][0];

    if (!inittab) {
        Pixel_F* _tab = (Pixel_F*)malloc(8 * INTER_TAB_SIZE * Pixel_F_sz);
        initInterTab1D(_tab);
        for (int i = 0; i < INTER_TAB_SIZE; i++ )
            for (int j = 0; j < INTER_TAB_SIZE; j++, tab += ksize * ksize, itab += ksize * ksize ) {
                int isum = 0;
                NNDeltaTab_i[i * INTER_TAB_SIZE + j][0] = j < INTER_TAB_SIZE / 2;
                NNDeltaTab_i[i * INTER_TAB_SIZE + j][1] = i < INTER_TAB_SIZE / 2;

                for (int k1 = 0; k1 < ksize; k1++ ) {
                    Pixel_F vy = _tab[i * ksize + k1];
                    for (int k2 = 0; k2 < ksize; k2++ ) {
                        Pixel_F v = vy * _tab[j * ksize + k2];
                        tab[k1 * ksize + k2] = v;
                        isum += itab[k1 * ksize + k2] = saturate_cast_to_short(v * INTER_REMAP_COEF_SCALE);
                    }
                }

                if( isum != INTER_REMAP_COEF_SCALE ) {
                    int diff = isum - INTER_REMAP_COEF_SCALE;
                    int Mk1    = 1,
                        Mk2    = 1,
                        mk1    = 1,
                        mk2    = 1;

                    for (int k1 = 1; k1 < 3; k1++ )
                        for (int k2 = 1; k2 < 3; k2++ ) {
                            if( itab[k1 * ksize + k2] < itab[mk1 * ksize + mk2] ) {
                                mk1 = k1;
                                mk2 = k2;
                            }
                            else if( itab[k1 * ksize + k2] > itab[Mk1 * ksize + Mk2] ) {
                                Mk1 = k1;
                                Mk2 = k2;
                            }
                        }

                    if( diff < 0 ) {
                        itab[Mk1 * ksize + Mk2] = (short)(itab[Mk1 * ksize + Mk2] - diff);
                    } else {
                        itab[mk1 * ksize + mk2] = (short)(itab[mk1 * ksize + mk2] - diff);
                    }
                }
            }

        itab   -= INTER_TAB_SIZE2 * ksize * ksize;
        inittab = true;
        free(_tab);
    }
    return (const void*)itab;
}

static Pixel_8 OpFixedPtCast(int val){
     return saturate_cast_to_Pixel_8((val + (1 << (INTER_REMAP_COEF_BITS-1))) >> INTER_REMAP_COEF_BITS);
}

// Functions for Planar8 warp.

static void remapBilinear_8( const vImage_Buffer *_src, const vImage_Buffer *_dst, const vImage_Buffer *_xy,
                           const vImage_Buffer *_fxy, const void* _wtab,
                           vImage_Flags borderType, Pixel_8 _borderValue ) {
    int ssize_height = _src->height;
    int ssize_width  = _src->width;
    int dsize_height = _dst->height;
    int dsize_width  = _dst->width;

    const short* wtab = (const short*)_wtab;
    const Pixel_8* S0 = (const Pixel_8*)_src->data;
    size_t sstep = _src->rowBytes / sizeof(S0[0]);

    unsigned width1  = max(ssize_width - 1, 0),
             height1 = max(ssize_height - 1, 0);

    for (int dy = 0; dy < dsize_height; dy++) {
        Pixel_8* D        =                ((Pixel_8*)_dst->data + _dst->rowBytes * dy);
        const short* XY   = (const short* )((Pixel_8*)_xy->data  + _xy->rowBytes  * dy);
        const ushort* FXY = (const ushort*)((Pixel_8*)_fxy->data + _fxy->rowBytes * dy);
        int X0 = 0;
        bool prevInlier = false;

        for (int dx = 0; dx <= dsize_width; dx++ ) {
            bool curInlier = dx < dsize_width ?
                (unsigned)XY[dx * 2]     < width1 &&
                (unsigned)XY[dx * 2 + 1] < height1 : !prevInlier;

            if (curInlier == prevInlier)   continue;

            int X1 = dx;
            dx = X0;
            X0 = X1;
            prevInlier = curInlier;

            if ( !curInlier ) {
                for ( ; dx < X1; dx++, D++ ) {
                    int sx = XY[dx * 2], sy = XY[dx * 2 + 1];
                    const short* w = wtab + FXY[dx] * 4;
                    const Pixel_8* S = S0 + sy * sstep + sx;
                    *D = OpFixedPtCast(int(S[0] * w[0] + S[1] * w[1] + S[sstep] * w[2] + S[sstep + 1] * w[3]));
                }

            } else {
                for (; dx < X1; dx++, D++ )   {
                    int sx = XY[dx * 2], sy = XY[dx * 2 + 1];
                    if( borderType == kvImageBackgroundColorFill &&
                        (sx >= ssize_width  || sx + 1 < 0 ||
                         sy >= ssize_height || sy + 1 < 0) ) {
                        D[0] = _borderValue;
                    } else {
                        int sx0, sx1, sy0, sy1;
                        Pixel_8 v0, v1, v2, v3;
                        const short* w = wtab + FXY[dx] * 4;
                        sx0 = clip(sx,     0, ssize_width);
                        sx1 = clip(sx + 1, 0, ssize_width);
                        sy0 = clip(sy,     0, ssize_height);
                        sy1 = clip(sy + 1, 0, ssize_height);
                        v0 = S0[sy0 * sstep + sx0];
                        v1 = S0[sy0 * sstep + sx1];
                        v2 = S0[sy1 * sstep + sx0];
                        v3 = S0[sy1 * sstep + sx1];
                        D[0] = OpFixedPtCast(int(v0 * w[0] + v1 * w[1] + v2 * w[2] + v3 * w[3]));
                    }
                }

            }
        }
    }
}

static void remap_8( const vImage_Buffer *src, const vImage_Buffer *dst,
               const vImage_Buffer *map1, const vImage_Buffer *map2,
               vImage_Flags borderType, Pixel_8 borderValue ) {

    assert(map1->data != NULL);
    assert(map2->data != NULL);
    assert(map1->height == map2->height);
    assert(map1->width  == map2->width);

    const void* ctab = 0;
    ctab = initInterTab2D();

    int brows0 = min(         128, dst->height);
    int bcols0 = min(16384/brows0, dst->width ); // 16384 = 1 << 14
        brows0 = min(16384/bcols0, dst->height);

    for(size_t y = 0; y < dst->height; y += brows0 ) {
        for(size_t x = 0; x < dst->width; x += bcols0 ) {
            int brows = min(brows0, dst->height - y);
            int bcols = min(bcols0, dst->width  - x);

            vImage_Buffer bufxy;       // (brows0, bcols0, CV_16SC2)
            bufxy.height   = brows;
            bufxy.width    = bcols;
            bufxy.rowBytes = map1->rowBytes;
            bufxy.data     = ((Pixel_8*)map1->data + map1->rowBytes * y) + 2 * sizeof(short) * x;

            vImage_Buffer  bufa;      // (brows0, bcols0, CV_16UC1)
            bufa.height    = brows;
            bufa.width     = bcols;
            bufa.rowBytes  = map2->rowBytes;
            bufa.data      = (Pixel_8*)map2->data + map2->rowBytes * y + x;

            vImage_Buffer dpart;
            dpart.height   = brows;
            dpart.width    = bcols;
            dpart.rowBytes = dst->rowBytes;
            dpart.data     = (Pixel_8*)dst->data + dpart.rowBytes * y + x;

            remapBilinear_8(src, &dpart, &bufxy, &bufa, ctab, borderType, borderValue);
        }
    }
}

static void invoker_8(const vImage_Buffer *src, const vImage_Buffer *dst, vImage_Flags borderType,
             Pixel_8 borderValue, int *adelta, int *bdelta, double *M)
{
    const int BLOCK_SZ       = 64,
              AB_BITS        = 10,
              AB_SCALE       = 1 << AB_BITS,
              round_delta    = AB_SCALE / INTER_TAB_SIZE / 2;

    short XY[BLOCK_SZ * BLOCK_SZ * 2],
           A[BLOCK_SZ * BLOCK_SZ];

    int bh0 = min(BLOCK_SZ / 2, dst->height);
    int bw0 = min(BLOCK_SZ * BLOCK_SZ / bh0, dst->width);
        bh0 = min(BLOCK_SZ * BLOCK_SZ / bw0, dst->height);

    for (int y = 0; y < (int)dst->height; y += bh0 ) {
        for (int x = 0; x < (int)dst->width; x += bw0 ) {
            int bw = min(bw0, dst->width  - x);
            int bh = min(bh0, dst->height - y);

            for (int y1 = 0; y1 < bh; y1++ ) {
                short* xy = XY + y1 * bw * 2;
                int X0 = saturate_cast_to_int((M[1] * (y + y1) + M[2]) * AB_SCALE) + round_delta;
                int Y0 = saturate_cast_to_int((M[4] * (y + y1) + M[5]) * AB_SCALE) + round_delta;

                short* alpha = A + y1 * bw;
                for (int x1 = 0; x1 < bw; x1++) {
                    int X = (X0 + adelta[x + x1]) >> (AB_BITS - INTER_BITS);
                    int Y = (Y0 + bdelta[x + x1]) >> (AB_BITS - INTER_BITS);
                    xy[x1 * 2]     = saturate_cast_to_short(X >> INTER_BITS);
                    xy[x1 * 2 + 1] = saturate_cast_to_short(Y >> INTER_BITS);
                    alpha[x1] = (short)((Y & (INTER_TAB_SIZE - 1)) * INTER_TAB_SIZE +
                            (X & (INTER_TAB_SIZE - 1)));
                }
            }

            vImage_Buffer _XY;
            _XY.height = bh;
            _XY.width  = bw;
            _XY.rowBytes = 2 * sizeof(short) * _XY.width;
            _XY.data = (void*)XY;

            vImage_Buffer dpart;
            dpart.height   = bh;
            dpart.width    = bw;
            dpart.rowBytes = dst->rowBytes;
            dpart.data     = (Pixel_8*)dst->data + dpart.rowBytes * y + x;

            vImage_Buffer _matA;
            _matA.height   = bh;
            _matA.width    = bw;
            _matA.rowBytes = _matA.width * sizeof(short);
            _matA.data     = (void*)A;

            remap_8( src, &dpart, &_XY, &_matA, borderType, borderValue );
        }
    }
}

void warpAffine_Planar8(const vImage_Buffer  *src,
                        const vImage_Buffer  *dest,
                        Pixel_D* M,
                        vImage_Flags borderType,
                        Pixel_8 borderValue)
{
    Pixel_D D = M[0] * M[4] - M[1] * M[3];
    D = D != 0 ? 1./D : 0;

    M[0]  = M[4] * D;
    M[1] *= -D;
    M[3] *= -D;
    M[4]  = M[0] * D;

    Pixel_D M5 = -M[3] * M[2] - M[4] * M[5];
    M[2] =       -M[0] * M[2] - M[1] * M[5];
    M[5] = M5;

    int *adelta = (int*)malloc(dest->width * 2 * sizeof(int)),
        *bdelta = adelta + dest->width;
    const int AB_BITS = max(10, INTER_BITS);
    const int AB_SCALE = 1 << AB_BITS;

    for( int x = 0; x < (int)dest->width; x++ ) {
        adelta[x] = saturate_cast_to_int(M[0] * x * AB_SCALE);
        bdelta[x] = saturate_cast_to_int(M[3] * x * AB_SCALE);
    }

    invoker_8(src, dest, borderType, borderValue, adelta, bdelta, M);
    free(adelta);
}

// Functions for ARGB8888 warp.

static void remapBilinear_8888( const vImage_Buffer* _src, const vImage_Buffer* _dst, const vImage_Buffer *_xy,
                           const vImage_Buffer* _fxy, const void* _wtab,
                           vImage_Flags borderType, Pixel_8888 _borderValue ){

    int ssize_height = _src->height,
        ssize_width  = _src->width;

    int dsize_height = _dst->height,
        dsize_width  = _dst->width;

    int cn = 4;
    const short* wtab = (const short*)_wtab;
    const Pixel_8* S0 = (const Pixel_8*)_src->data;
    size_t sstep = _src->rowBytes/sizeof(S0[0]);


    unsigned width1 = max(ssize_width-1, 0), height1 = max(ssize_height-1, 0);

    for(int dy = 0; dy < dsize_height; dy++ )    {
        Pixel_8* D        =               ((Pixel_8*)_dst->data  + _dst->rowBytes * dy);
        const short* XY   = (const short* )((Pixel_8*)_xy->data  + _xy->rowBytes  * dy);
        const ushort* FXY = (const ushort*)((Pixel_8*)_fxy->data + _fxy->rowBytes * dy);
        int X0 = 0;
        bool prevInlier = false;

        for(int dx = 0; dx <= dsize_width; dx++ )
        {
            bool curInlier = dx < dsize_width ?
                (unsigned)XY[dx*2] < width1 &&
                (unsigned)XY[dx*2+1] < height1 : !prevInlier;
            if( curInlier == prevInlier )
                continue;

            int X1 = dx;
            dx = X0;
            X0 = X1;
            prevInlier = curInlier;

            if( !curInlier ) {

                for( ; dx < X1; dx++, D += 4 ) {
                    int sx = XY[dx*2], sy = XY[dx*2+1];
                    const short* w = wtab + FXY[dx]*4;
                    const Pixel_8* S = S0 + sy*sstep + sx*4;
                    const int t0 = S[0]*w[0] + S[4]*w[1] + S[sstep+0]*w[2] + S[sstep+4]*w[3];
                    const int t1 = S[1]*w[0] + S[5]*w[1] + S[sstep+1]*w[2] + S[sstep+5]*w[3];
                    const int t2 = S[2]*w[0] + S[6]*w[1] + S[sstep+2]*w[2] + S[sstep+6]*w[3];
                    const int t3 = S[3]*w[0] + S[7]*w[1] + S[sstep+3]*w[2] + S[sstep+7]*w[3];
                    D[0] = OpFixedPtCast(t0);
                    D[1] = OpFixedPtCast(t1);
                    D[2] = OpFixedPtCast(t2);
                    D[3] = OpFixedPtCast(t3);
                }
            }
            else {
                    for( ; dx < X1; dx++, D += cn )  {
                        int sx = XY[dx*2], sy = XY[dx*2+1], k;
                        if( borderType == kvImageBackgroundColorFill &&
                            (sx >= ssize_width || sx+1 < 0 ||
                             sy >= ssize_height || sy+1 < 0) ) {
                            for( k = 0; k < cn; k++ ) {
                                D[k] = _borderValue[k];
                            }
                        } else {
                            int sx0, sx1, sy0, sy1;
                            const Pixel_8 *v0, *v1, *v2, *v3;
                            const short* w = wtab + FXY[dx]*4;
                            sx0 = clip(sx    , 0, ssize_width);
                            sx1 = clip(sx + 1, 0, ssize_width);
                            sy0 = clip(sy,     0, ssize_height);
                            sy1 = clip(sy+1,   0, ssize_height);
                            v0 = S0 + sy0 * sstep + sx0 * cn;
                            v1 = S0 + sy0 * sstep + sx1 * cn;
                            v2 = S0 + sy1 * sstep + sx0 * cn;
                            v3 = S0 + sy1 * sstep + sx1 * cn;
                            for( k = 0; k < cn; k++ )
                                D[k] = OpFixedPtCast(int(v0[k]*w[0] + v1[k]*w[1] + v2[k]*w[2] + v3[k]*w[3]));
                        }
                    }
            }
        }
    }
}

static void remap_8888( const vImage_Buffer *src,  const vImage_Buffer *dst,
                 const vImage_Buffer *map1, const vImage_Buffer *map2,
                 vImage_Flags borderType,   Pixel_8888 borderValue ){

    assert(map1->data != NULL);
    assert(map2->data != NULL);
    assert(map1->height == map2->height);
    assert(map1->width  == map2->width);

    const void* ctab = 0;
    ctab = initInterTab2D();

    int brows0 = min(         128, dst->height);
    int bcols0 = min(16384/brows0, dst->width ); // 16384 = 1 << 14
        brows0 = min(16384/bcols0, dst->height);

    for(size_t y = 0; y < dst->height; y += brows0 ) {
         for(size_t x = 0; x < dst->width; x += bcols0 ) {
            int brows = min(brows0, dst->height - y);
            int bcols = min(bcols0, dst->width  - x);

            vImage_Buffer dpart;
            dpart.height   = brows;
            dpart.width    = bcols;
            dpart.rowBytes = dst->rowBytes;
            dpart.data     = (Pixel_8*)dst->data + dpart.rowBytes * y + x;

            vImage_Buffer bufxy;       // (brows0, bcols0, CV_16SC2)
            bufxy.height   = brows;
            bufxy.width    = bcols;
            bufxy.rowBytes = map1->rowBytes;
            bufxy.data     = ((Pixel_8*)map1->data + map1->rowBytes * y) + 2 * sizeof(short) * x;

            vImage_Buffer  bufa;      // (brows0, bcols0, CV_16UC1)
            bufa.height    = brows;
            bufa.width     = bcols;
            bufa.rowBytes  = map2->rowBytes;
            bufa.data      = (Pixel_8*)map2->data + map2->rowBytes * y + x;

            remapBilinear_8888(src, &dpart, &bufxy, &bufa, ctab, borderType, borderValue);
        }
    }
}

static void invoker_8888(const vImage_Buffer *src, const vImage_Buffer *dst, vImage_Flags borderType,
             Pixel_8888 borderValue, int *adelta, int *bdelta, double *M)
{
    const int BLOCK_SZ       = 64,
              AB_BITS        = 10,
              AB_SCALE       = 1 << AB_BITS,
              round_delta    = AB_SCALE / INTER_TAB_SIZE / 2;

    short XY[BLOCK_SZ * BLOCK_SZ * 2],
           A[BLOCK_SZ * BLOCK_SZ];

    int bh0 = min(BLOCK_SZ / 2, dst->height);
    int bw0 = min(BLOCK_SZ * BLOCK_SZ / bh0, dst->width);
        bh0 = min(BLOCK_SZ * BLOCK_SZ / bw0, dst->height);

    for (int y = 0; y < (int)dst->height; y += bh0 ) {
        for (int x = 0; x < (int)dst->width; x += bw0 ) {
            int bw = min(bw0, dst->width  - x);
            int bh = min(bh0, dst->height - y);

            for (int y1 = 0; y1 < bh; y1++ ) {
                short* xy = XY + y1 * bw * 2;
                int X0 = saturate_cast_to_int((M[1] * (y + y1) + M[2]) * AB_SCALE) + round_delta;
                int Y0 = saturate_cast_to_int((M[4] * (y + y1) + M[5]) * AB_SCALE) + round_delta;

                short* alpha = A + y1 * bw;
                for (int x1 = 0; x1 < bw; x1++) {
                    int X = (X0 + adelta[x + x1]) >> (AB_BITS - INTER_BITS);
                    int Y = (Y0 + bdelta[x + x1]) >> (AB_BITS - INTER_BITS);
                    xy[x1 * 2]     = saturate_cast_to_short(X >> INTER_BITS); // ?
                    xy[x1 * 2 + 1] = saturate_cast_to_short(Y >> INTER_BITS); // ?
                    alpha[x1] = (short)((Y & (INTER_TAB_SIZE - 1)) * INTER_TAB_SIZE +
                            (X & (INTER_TAB_SIZE - 1)));
                }
            }

            vImage_Buffer _XY;
            _XY.height = bh;
            _XY.width  = bw;
            _XY.rowBytes = 2 * sizeof(short) * _XY.width;
            _XY.data = (void*)XY;

            vImage_Buffer dpart;
            dpart.height   = bh;
            dpart.width    = bw;
            dpart.rowBytes = dst->rowBytes;
            dpart.data     = (Pixel_8*)dst->data + dpart.rowBytes * y + x * Pixel_8888_sz;

            vImage_Buffer _matA;
            _matA.height   = bh;
            _matA.width    = bw;
            _matA.rowBytes = _matA.width * sizeof(short);
            _matA.data     = (void*)A;

            remap_8888( src, &dpart, &_XY, &_matA, borderType, borderValue );
        }
    }
}

void warpAffine_ARGB8888(
    const vImage_Buffer *src,
    vImage_Buffer *dest,
    Pixel_D* M,
    vImage_Flags borderType,
    Pixel_8888 borderValue)
{
    Pixel_D D = M[0] * M[4] - M[1] * M[3];
    D = D != 0 ? 1./D : 0;

    M[0]  = M[4] * D;
    M[1] *= -D;
    M[3] *= -D;
    M[4]  = M[0] * D;

    Pixel_D M5 = -M[3] * M[2] - M[4] * M[5];
    M[2] =       -M[0] * M[2] - M[1] * M[5];
    M[5] = M5;

    int *adelta = (int*)malloc(dest->width * 2 * sizeof(int)),
        *bdelta = adelta + dest->width;
    const int AB_BITS  = 10;
    const int AB_SCALE = 1 << AB_BITS;

    for( int x = 0; x < (int)dest->width; x++ ) {
        adelta[x] = saturate_cast_to_int(M[0] * x * AB_SCALE);
        bdelta[x] = saturate_cast_to_int(M[3] * x * AB_SCALE);
    }

    invoker_8888(src, dest, borderType, borderValue, adelta, bdelta, M);
    free(adelta);
}
