//
//  GrabCut.cpp
//  FaceRecognation
//
//  Created by Lusine Abrahamyan on 6/15/16.
//  Copyright © 2016 Lusine. All rights reserved.
//

#include "GrabCut.hpp"
#include "Kmeans.hpp"
#include "Kmeans.cpp"
#include "gcgraph.h"
#include <limits>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <jni.h>
#include <android/bitmap.h>
#include <android/log.h>

#include "Accelerate/Accelerate.h"

#ifndef EFFECTS_USE_GCD
#define EFFECTS_USE_GCD 0
#endif


#if EFFECTS_USE_GCD
#include <dispatch/dispatch.h>
#endif

#define DISPATCH_QUEUE_PRIORITY_HIGH 2
#define DISPATCH_QUEUE_PRIORITY_DEFAULT 0
#define DISPATCH_QUEUE_PRIORITY_LOW (-2)
#define DISPATCH_QUEUE_PRIORITY_BACKGROUND INT16_MIN

#define ANDROID_LOG_TAG "libgrabcut"

struct Pairs {
	Point2I p1;
	Point2I p2;
};

/*
 This is implementation of image segmentation algorithm GrabCut described in
 "GrabCut — Interactive Foreground Extraction using Iterated Graph Cuts".
 Carsten Rother, Vladimir Kolmogorov, Andrew Blake.
 */

/*
 GMM - Gaussian Mixture Model
 */
class GMM {
public:
	static const int componentsCount = 5;

	GMM();

	double operator()(const Pixel color) const;

	double operator()(int ci, const Pixel color) const;

	int whichComponent(const Pixel color) const;

	void initLearning();

	void addSample(int ci, const Pixel color);

	void endLearning();

private:
	void calcInverseCovAndDeterm(int ci);

	double *coefs;
	double *mean;
	double *cov;

	double inverseCovs[componentsCount][3][3];
	double covDeterms[componentsCount];

	double sums[componentsCount][3];
	double prods[componentsCount][3][3];
	int sampleCounts[componentsCount];
	int totalSampleCount;
};


GMM::GMM() {
	const int modelSize = 3/*mean*/ + 9/*covariance*/ + 1/*component weight*/;

	coefs = (double *) malloc(sizeof(double) * modelSize * componentsCount);
	memset(coefs, 0, modelSize * componentsCount); // TODO

	mean = coefs + componentsCount;
	cov = mean + 3 * componentsCount;

	for (int ci = 0; ci < componentsCount; ci++)
		if (coefs[ci] > 0)
			calcInverseCovAndDeterm(ci);
}


double GMM::operator()(const Pixel color) const {
	double res = 0;
	for (int ci = 0; ci < componentsCount; ci++)
		res += coefs[ci] * (*this)(ci, color);
	return res;
}


double GMM::operator()(int ci, const Pixel color) const {
	double res = 0;
	if (coefs[ci] > 0) {
		double diff[] = {static_cast<double>(color.red), static_cast<double>(color.green),
						 static_cast<double>(color.blue)};
		double *m = mean + 3 * ci;
		diff[0] -= m[0];
		diff[1] -= m[1];
		diff[2] -= m[2];
		double mult = diff[0] * (diff[0] * inverseCovs[ci][0][0] + diff[1] * inverseCovs[ci][1][0] +
								 diff[2] * inverseCovs[ci][2][0])
					  + diff[1] *
						(diff[0] * inverseCovs[ci][0][1] + diff[1] * inverseCovs[ci][1][1] +
						 diff[2] * inverseCovs[ci][2][1])
					  + diff[2] *
						(diff[0] * inverseCovs[ci][0][2] + diff[1] * inverseCovs[ci][1][2] +
						 diff[2] * inverseCovs[ci][2][2]);
		res = 1.0f / sqrt(covDeterms[ci]) * exp(-0.5f * mult);
	}
	return res;
}


int GMM::whichComponent(Pixel color) const {
	int k = 0;
	double max = 0;

	for (int ci = 0; ci < componentsCount; ci++) {
		double p = (*this)(ci, color);
		if (p > max) {
			k = ci;
			max = p;
		}
	}
	return k;
}

void GMM::initLearning() {
	for (int ci = 0; ci < componentsCount; ci++) {
		sums[ci][0] = sums[ci][1] = sums[ci][2] = 0;
		prods[ci][0][0] = prods[ci][0][1] = prods[ci][0][2] = 0;
		prods[ci][1][0] = prods[ci][1][1] = prods[ci][1][2] = 0;
		prods[ci][2][0] = prods[ci][2][1] = prods[ci][2][2] = 0;
		sampleCounts[ci] = 0;
	}
	totalSampleCount = 0;
}


void GMM::addSample(int ci, const Pixel color) {

	sums[ci][0] += color.red;
	sums[ci][1] += color.green;
	sums[ci][2] += color.blue;
	prods[ci][0][0] += color.red * color.red;
	prods[ci][0][1] += color.red * color.green;
	prods[ci][0][2] += color.red * color.blue;
	prods[ci][1][0] += color.green * color.red;
	prods[ci][1][1] += color.green * color.green;
	prods[ci][1][2] += color.green * color.blue;
	prods[ci][2][0] += color.blue * color.red;
	prods[ci][2][1] += color.blue * color.green;
	prods[ci][2][2] += color.blue * color.blue;
	sampleCounts[ci]++;
	totalSampleCount++;
}

void GMM::endLearning() {
	const double variance = 0.01;
	for (int ci = 0; ci < componentsCount; ci++) {
		int n = sampleCounts[ci];
		if (n == 0)
			coefs[ci] = 0;
		else {
			coefs[ci] = (double) n / totalSampleCount;

			double *m = mean + 3 * ci;
			m[0] = sums[ci][0] / n;
			m[1] = sums[ci][1] / n;
			m[2] = sums[ci][2] / n;

			double *c = cov + 9 * ci;
			c[0] = prods[ci][0][0] / n - m[0] * m[0];
			c[1] = prods[ci][0][1] / n - m[0] * m[1];
			c[2] = prods[ci][0][2] / n - m[0] * m[2];
			c[3] = prods[ci][1][0] / n - m[1] * m[0];
			c[4] = prods[ci][1][1] / n - m[1] * m[1];
			c[5] = prods[ci][1][2] / n - m[1] * m[2];
			c[6] = prods[ci][2][0] / n - m[2] * m[0];
			c[7] = prods[ci][2][1] / n - m[2] * m[1];
			c[8] = prods[ci][2][2] / n - m[2] * m[2];

			double dtrm = c[0] * (c[4] * c[8] - c[5] * c[7]) - c[1] * (c[3] * c[8] - c[5] * c[6]) +
						  c[2] * (c[3] * c[7] - c[4] * c[6]);
			if (dtrm <= std::numeric_limits<double>::epsilon()) {
				// Adds the white noise to avoid singular covariance matrix.
				c[0] += variance;
				c[4] += variance;
				c[8] += variance;
			}

			calcInverseCovAndDeterm(ci);
		}
	}
}

void GMM::calcInverseCovAndDeterm(int ci) {
	if (coefs[ci] > 0) {
		double *c = cov + 9 * ci;
		double dtrm =
				covDeterms[ci] =
						c[0] * (c[4] * c[8] - c[5] * c[7]) - c[1] * (c[3] * c[8] - c[5] * c[6]) +
						c[2] * (c[3] * c[7] - c[4] * c[6]);
		inverseCovs[ci][0][0] = (c[4] * c[8] - c[5] * c[7]) / dtrm;
		inverseCovs[ci][1][0] = -(c[3] * c[8] - c[5] * c[6]) / dtrm;
		inverseCovs[ci][2][0] = (c[3] * c[7] - c[4] * c[6]) / dtrm;
		inverseCovs[ci][0][1] = -(c[1] * c[8] - c[2] * c[7]) / dtrm;
		inverseCovs[ci][1][1] = (c[0] * c[8] - c[2] * c[6]) / dtrm;
		inverseCovs[ci][2][1] = -(c[0] * c[7] - c[1] * c[6]) / dtrm;
		inverseCovs[ci][0][2] = (c[1] * c[5] - c[2] * c[4]) / dtrm;
		inverseCovs[ci][1][2] = -(c[0] * c[5] - c[2] * c[3]) / dtrm;
		inverseCovs[ci][2][2] = (c[0] * c[4] - c[1] * c[3]) / dtrm;
	}
}


Pixel getPixelAt(const Pixel_8 *const imageData, size_t rowBytes, size_t x, size_t y) {
	Pixel rgb;

	rgb.red = imageData[y * rowBytes + 4 * x + 1];
	rgb.green = imageData[y * rowBytes + 4 * x + 2];
	rgb.blue = imageData[y * rowBytes + 4 * x + 3];

	return rgb;
}

double dotDiff(Pixel first, Pixel second) {
	int diff[3] = {static_cast<int>(first.red - second.red),
				   static_cast<int>(first.green - second.green),
				   static_cast<int>(first.blue - second.blue)};
	double selfDotProduct = pow(diff[0], 2) + pow(diff[0], 2) + pow(diff[2], 2);
	return selfDotProduct;
}

static double calcBeta(const vImage_Buffer &sourceImage) {
	double beta = 0;
	auto rb = sourceImage.rowBytes;

	for (int y = 0; y < sourceImage.height; y++) {
		for (int x = 0; x < sourceImage.width; x++) {
			Pixel color = getPixelAt((Pixel_8 *) sourceImage.data, rb, x, y);

			if (x > 0) // left
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, rb, x - 1, y);
				beta += dotDiff(color, compare);
			}
			if (y > 0 && x > 0) // upleft
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, rb, x - 1, y - 1);
				beta += dotDiff(color, compare);
			}
			if (y > 0) // up
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, rb, x, y - 1);
				beta += dotDiff(color, compare);
			}
			if (y > 0 && x < sourceImage.width - 1) // upright
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, rb, x + 1, y - 1);
				beta += dotDiff(color, compare);
			}
		}
	}
	if (beta <= std::numeric_limits<double>::epsilon())
		beta = 0;
	else
		beta = 1.f / (2 * beta /
					  (4 * sourceImage.height * sourceImage.width - 3 * sourceImage.width -
					   3 * sourceImage.height + 2));

	return beta;
}


double computeWeight(Pixel first, Pixel second, double beta, double gamma) {

	int dif[3];
	dif[0] = first.red - second.red;
	dif[1] = first.green - second.green;
	dif[2] = first.blue - second.blue;

	double squareDiff = pow(dif[0], 2) + pow(dif[1], 2) + pow(dif[2], 2);
	double weight = gamma * exp(-beta * squareDiff);
	return weight;
}

/*
 Computing weights of each pixel with their neighbours
 */
static void calcNWeights(double beta, double gamma, const vImage_Buffer &sourceImage,
						 vImage_Buffer &leftWB, vImage_Buffer &upleftWB, vImage_Buffer &upWB,
						 vImage_Buffer &uprightWB) {
	const double gammaDivSqrt2 = gamma / std::sqrt(2.0f);
	auto rb = leftWB.rowBytes;

	for (int y = 0; y < sourceImage.height; y++) {
		for (int x = 0; x < sourceImage.width; x++) {
			Pixel currentPixel = getPixelAt((Pixel_8 *) sourceImage.data, sourceImage.rowBytes, x,
											y);

			if (x - 1 >= 0) // left
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, sourceImage.rowBytes,
										   x - 1, y);
				((float *) leftWB.data)[y * rb + x] = computeWeight(currentPixel, compare, beta,
																	gamma);

			}
			else {
				((float *) leftWB.data)[y * rb + x] = 0;

			}
			if (x - 1 >= 0 && y - 1 >= 0) // upleft
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, sourceImage.rowBytes,
										   x - 1, y - 1);
				((float *) upleftWB.data)[y * rb + x] = computeWeight(currentPixel, compare, beta,
																	  gamma);

			}
			else {
				((float *) upleftWB.data)[y * rb + x] = 0;

			}
			if (y - 1 >= 0) // up
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, sourceImage.rowBytes, x,
										   y - 1);
				((float *) upWB.data)[y * rb + x] = computeWeight(currentPixel, compare, beta,
																  gamma);

			}
			else {
				((float *) upWB.data)[y * rb + x] = 0;

			}
			if (x + 1 < sourceImage.width && y - 1 >= 0) // upright
			{
				Pixel compare = getPixelAt((Pixel_8 *) sourceImage.data, sourceImage.rowBytes,
										   x + 1, y - 1);
				((float *) uprightWB.data)[y * rb + x] = computeWeight(currentPixel, compare, beta,
																	   gamma);

			}
			else {
				((float *) uprightWB.data)[y * rb + x] = 0;

			}
		}
	}
}


/*
 Initialize GMM background and foreground models using kmeans algorithm.
 */
static int initGMMs(GMM &bgdGMM, GMM &fgdGMM, const vImage_Buffer &sourceImage,
					const vImage_Buffer &binaryMask) {
	std::vector<Pixel> bgdSamplesNew, fgSamplesNew;
	auto rb = sourceImage.rowBytes;
	Point2I p;
	auto pathAmount = 0;
	for (p.y = 0; p.y < sourceImage.height; p.y++) {
		for (p.x = 0; p.x < sourceImage.width; p.x++) {
			if (((Pixel_8 *) binaryMask.data)[p.y * binaryMask.rowBytes + p.x] == GC_BGD) {

				Pixel pix = getPixelAt((Pixel_8 *) sourceImage.data, rb, p.x, p.y);

				bgdSamplesNew.push_back(pix);

			} else {// GC_FGD | GC_PR_FGD

				Pixel pix = getPixelAt((Pixel_8 *) sourceImage.data, rb, p.x, p.y);

				fgSamplesNew.push_back(pix);
				pathAmount++;
			}
		}
	}

	if (fgSamplesNew.empty() || bgdSamplesNew.empty()) {

		return -1;
	}

	int *bgLabels = (int *) malloc(sizeof(int) * bgdSamplesNew.size());
	int *fgLabels = (int *) malloc(sizeof(int) * fgSamplesNew.size());

	kmeans((Pixel *) bgdSamplesNew.data(), GMM::componentsCount, bgdSamplesNew.size(), bgLabels);
	kmeans((Pixel *) fgSamplesNew.data(), GMM::componentsCount, fgSamplesNew.size(), fgLabels);

	bgdGMM.initLearning();
	for (int i = 0; i < (int) bgdSamplesNew.size(); i++) {
		bgdGMM.addSample(bgLabels[i], bgdSamplesNew[i]);
	}
	bgdGMM.endLearning();

	fgdGMM.initLearning();
	for (int i = 0; i < (int) fgSamplesNew.size(); i++) {
		fgdGMM.addSample(fgLabels[i], fgSamplesNew[i]);
	}
	fgdGMM.endLearning();

	free(bgLabels);
	bgLabels = nullptr;
	free(fgLabels);
	fgLabels = nullptr;

	return pathAmount;
}


struct CompIdxContext {

	const vImage_Buffer *source;
	vImage_Buffer *mask;
	vImage_Buffer *compIdx;
	GMM fgGMM;
	GMM bgGMM;
};

void assignGMMsComponents_parallel(void *context, size_t iter) {

	CompIdxContext *cnt = (CompIdxContext *) context;
	Pixel_8 *maskBuffer = (Pixel_8 *) cnt->mask->data;
	Pixel_8 *compIdxBuffer = (Pixel_8 *) cnt->compIdx->data;

	auto rb = cnt->source->rowBytes;
	Point2I p;
	p.y = iter;
	for (p.x = 0; p.x < cnt->source->width; p.x++) {
		Pixel color = getPixelAt((Pixel_8 *) cnt->source->data, rb, p.x, p.y);

		compIdxBuffer[p.y * cnt->compIdx->rowBytes + p.x] =
				maskBuffer[p.y * cnt->mask->rowBytes + p.x] == GC_BGD ?
				cnt->bgGMM.whichComponent(color) : cnt->fgGMM.whichComponent(color);
	}

}

void assignGMMsComponents(const GMM &bgdGMM, const GMM &fgdGMM, vImage_Buffer &compIdxs,
						  const vImage_Buffer &sourceImage, vImage_Buffer &binaryMask) {

	CompIdxContext context;
	context.bgGMM = bgdGMM;
	context.fgGMM = fgdGMM;
	context.source = &sourceImage;
	context.mask = &binaryMask;
	context.compIdx = &compIdxs;

#if EFFECTS_USE_GCD
	dispatch_apply_f(sourceImage.height, dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), &context, assignGMMsComponents_parallel);
#else
#pragma omp parallel for
	for (int iter = 0; iter < sourceImage.height; ++iter)
		assignGMMsComponents_parallel(&context, iter);
#endif

}

/*
 Learn GMMs parameters.
 */
static void learnGMMs(const vImage_Buffer &compIdxs, GMM &bgdGMM, GMM &fgdGMM,
					  const vImage_Buffer &sourceImage, vImage_Buffer &binaryMask) {
	bgdGMM.initLearning();
	fgdGMM.initLearning();
	Point2I p;
	Pixel_8 *maskBuffer = (Pixel_8 *) binaryMask.data;
	auto rb = sourceImage.rowBytes;
	for (int ci = 0; ci < GMM::componentsCount; ci++) {
		for (p.y = 0; p.y < sourceImage.height; p.y++) {
			for (p.x = 0; p.x < sourceImage.width; p.x++) {
				if (((Pixel_8 *) compIdxs.data)[p.y * compIdxs.rowBytes + p.x] == ci) {
					Pixel color = getPixelAt((Pixel_8 *) sourceImage.data, rb, p.x, p.y);

					if (maskBuffer[p.y * binaryMask.rowBytes + p.x] == GC_BGD)
						bgdGMM.addSample(ci, color);
					else
						fgdGMM.addSample(ci, color);
				}
			}
		}
	}
	bgdGMM.endLearning();
	fgdGMM.endLearning();
}

/*
 Construct graph with computed weights
 */
void constructGCGraph(const GMM &bgdGMM, const GMM &fgdGMM, double lambda, GCGraph<double> &graph,
					  const vImage_Buffer &sourceImage, const vImage_Buffer &binaryMask,
					  const vImage_Buffer &leftWB, const vImage_Buffer &upleftWB,
					  const vImage_Buffer &upWB, const vImage_Buffer &uprightWB) {
	auto vtxCount = sourceImage.width * sourceImage.height,
			edgeCount = 2 * (4 * sourceImage.width * sourceImage.height -
							 3 * (sourceImage.width + sourceImage.height) + 2);
	graph.create(vtxCount, edgeCount);
	Point2I p;
	Pixel_8 *maskBuffer = (Pixel_8 *) binaryMask.data;
	auto rb = binaryMask.rowBytes;

	for (p.y = 0; p.y < sourceImage.height; p.y++) {
		for (p.x = 0; p.x < sourceImage.width; p.x++) {
			// add node
			int vtxIdx = graph.addVtx();
			Pixel color = getPixelAt((Pixel_8 *) sourceImage.data, sourceImage.rowBytes, p.x, p.y);

			// set t-weights
			double fromSource, toSink;
			if (maskBuffer[p.y * rb + p.x] == GC_PR_FGD) {
				fromSource = -log(bgdGMM(color));
				toSink = -log(fgdGMM(color));
			}
			else if (maskBuffer[p.y * rb + p.x] == GC_BGD) {
				fromSource = 0;
				toSink = lambda;
			}
			else // GC_FGD
			{
				fromSource = lambda;
				toSink = 0;
			}
			graph.addTermWeights(vtxIdx, fromSource, toSink);

			// set n-weights
			if (p.x > 0) {
				float w = ((float *) leftWB.data)[p.y * rb + p.x];
				graph.addEdges(vtxIdx, vtxIdx - 1, w, w);
			}
			if (p.x > 0 && p.y > 0) {
				float w = ((float *) upleftWB.data)[p.y * rb + p.x];
				graph.addEdges(vtxIdx, vtxIdx - sourceImage.width - 1, w, w);
			}
			if (p.y > 0) {
				float w = ((float *) upWB.data)[p.y * rb + p.x];
				graph.addEdges(vtxIdx, vtxIdx - sourceImage.width, w, w);
			}
			if (p.x < sourceImage.width - 1 && p.y > 0) {
				float w = ((float *) uprightWB.data)[p.y * rb + p.x];
				graph.addEdges(vtxIdx, vtxIdx - sourceImage.width + 1, w, w);
			}
		}
	}
}

/*
 Estimate segmentation using MaxFlow algorithm
 */
static int estimateSegmentation(GCGraph<double> &graph, vImage_Buffer &binaryMask) {
	vImage_Buffer mask = {malloc(sizeof(Pixel_8) * binaryMask.rowBytes * binaryMask.height),
						  binaryMask.height, binaryMask.width, binaryMask.rowBytes};
	memcpy((Pixel_8 *) mask.data, (Pixel_8 *) binaryMask.data,
		   binaryMask.height * binaryMask.rowBytes);

	int maskAmount = 0;
	graph.maxFlow();

	Point2I p;
	Pixel_8 *maskBuffer = (Pixel_8 *) binaryMask.data;
	memset(maskBuffer, 0, binaryMask.rowBytes * binaryMask.height);

	auto rb = binaryMask.rowBytes;
	for (p.y = 0; p.y < binaryMask.height; p.y++) {
		for (p.x = 0; p.x < binaryMask.width; p.x++) {
			if (graph.inSourceSegment((p.y * rb + p.x) /*vertex index*/ )) {
				maskBuffer[p.y * rb + p.x] = GC_PR_FGD;
				++maskAmount;
			} else if (p.x > 1 && graph.inSourceSegment(
					p.y * rb + (p.x - 1) /*check if blob is in mask bounds*/ )) {
				for (auto i = p.x; i < binaryMask.width; i++) {
					if (maskBuffer[(p.y - 1) * rb + i] == GC_PR_FGD) {

						if (graph.inSourceSegment(p.y * rb + i) &&
							((Pixel_8 *) mask.data)[p.y * rb + i] == GC_PR_FGD) {
							//fill blob
							for (auto m = p.x; m <= i; m++) {
								maskBuffer[p.y * rb + m] = GC_PR_FGD;
								++maskAmount;
							}
							break;
						}
						if (!graph.inSourceSegment(p.y * rb + i) &&
							((Pixel_8 *) mask.data)[p.y * rb + i] == GC_BGD) {
							break;
						}
					} else {
						break;
					}
				}
			}
		}
	}

	__android_log_print(ANDROID_LOG_ERROR, ANDROID_LOG_TAG, "found result: %d", maskAmount);

	free(mask.data);
	mask.data = nullptr;
	return maskAmount;
}


/*
 Estimate segmentation using MaxFlow algorithm in refine mode
 */
static int refineSegmentation(GCGraph<double> &graph, vImage_Buffer &resultMask,
							  vImage_Buffer &binaryMask, int mode) {
	graph.maxFlow();

	Point2I p;
	Pixel_8 *maskBuffer = (Pixel_8 *) resultMask.data;
	auto rb = binaryMask.rowBytes;
	auto resultAmount = 0;
	for (p.y = 0; p.y < binaryMask.height; p.y++) {
		for (p.x = 0; p.x < binaryMask.width; p.x++) {
			if (graph.inSourceSegment(p.y * rb + p.x /*vertex index*/) &&
				((Pixel_8 *) binaryMask.data)[p.y * binaryMask.rowBytes + p.x] != mode) {
				maskBuffer[p.y * rb + p.x] = GC_PR_FGD;
				resultAmount++;
			} else {
				maskBuffer[p.y * rb + p.x] = GC_BGD;
			}
		}
	}
	return resultAmount;
}

/*
 binaryBuffer - 1 channel plana mask to smooth edges
 smoothAmount - smoothing parameter in range 0 - 100
 */
void smoothBuffer(vImage_Buffer &binaryBuffer, unsigned smoothAmount) {

	auto maxSmoothing = 17;
	smoothAmount = ((float) smoothAmount / 100) * maxSmoothing;
	smoothAmount = (int) (smoothAmount) % 2 == 0 ? smoothAmount - 1 > 0 ? smoothAmount - 1 : 0
												 : smoothAmount;

	if (smoothAmount != 0) {
		vImageBoxConvolve_Planar8(&binaryBuffer,
								  &binaryBuffer,
								  NULL,
								  0,
								  0,
								  smoothAmount,
								  smoothAmount,
								  0,
								  kvImageBackgroundColorFill);
	}
}

/*
 magic lasso main function.
 Max side of input image is 1000.
 sourceImage - ARGB buffer
 binaryMask - Planar8 buffer
 progress - algorithm state in percentages
 */
bool letTheMagicBegin(const vImage_Buffer &sourceImage, vImage_Buffer &binaryMask, int &progress) {
	if (sourceImage.width != binaryMask.width || sourceImage.height != binaryMask.height) {
		return false;
	}

	progress = 0;
	auto weightsBufferSize = sizeof(float *) * sourceImage.width * sourceImage.height;

	GMM bgdGMM, fgdGMM;

	auto pathAmount = initGMMs(bgdGMM, fgdGMM, sourceImage, binaryMask);
	if (pathAmount == -1) {
		return false;
	}

	progress = 5;
	const double gamma = 50;
	const double lambda = 9 * gamma;
	const double beta = calcBeta(sourceImage);

	vImage_Buffer upWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
						  sourceImage.width};
	vImage_Buffer uprightWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
							   sourceImage.width};
	vImage_Buffer leftWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
							sourceImage.width};
	vImage_Buffer upleftWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
							  sourceImage.width};

	calcNWeights(beta, gamma, sourceImage, leftWB, upleftWB, upWB, uprightWB);
	progress = 10;

	vImage_Buffer compIdxs = {malloc(sizeof(Pixel_8 *) * sourceImage.width * sourceImage.height),
							  sourceImage.height, sourceImage.width, sourceImage.width};
	GCGraph<double> graph;
	assignGMMsComponents(bgdGMM, fgdGMM, compIdxs, sourceImage, binaryMask);
	progress = 20;

	learnGMMs(compIdxs, bgdGMM, fgdGMM, sourceImage, binaryMask);
	progress = 25;
	constructGCGraph(bgdGMM, fgdGMM, lambda, graph, sourceImage, binaryMask, leftWB, upleftWB, upWB,
					 uprightWB);
	progress = 50;

	progress = 100;
	free(compIdxs.data);
	compIdxs.data = nullptr;
	free(leftWB.data);
	leftWB.data = nullptr;
	free(upleftWB.data);
	upleftWB.data = nullptr;
	free(upWB.data);
	upWB.data = nullptr;
	free(uprightWB.data);
	uprightWB.data = nullptr;

	int maskAmount = estimateSegmentation(graph, binaryMask);

	return (float) maskAmount/pathAmount > 0.3f;
}

/*
In mask refine mode merging binaryMask with output of algorithm
 */
static void refineInput(vImage_Buffer &resultMask, vImage_Buffer &binaryMask, int mode) {

	Pixel_8 *buffer = (Pixel_8 *) resultMask.data;
	auto rb = resultMask.rowBytes;

	for (int y = 0; y < binaryMask.height; y++) {
		for (int x = 0; x < binaryMask.width; x++) {

			if (buffer[y * rb + x] == GC_PR_FGD) {
				((Pixel_8 *) binaryMask.data)[y * binaryMask.rowBytes + x] = mode;
			}
		}
	}
}

/*
 In mask refine mode merging binaryMask with input of user selection
 */
static void returnInput(vImage_Buffer &refineMask, vImage_Buffer &binaryMask, int mode) {

	for (int y = 0; y < refineMask.height; y++) {
		for (int x = 0; x < refineMask.width; x++) {

			if (((Pixel_8 *) refineMask.data)[y * refineMask.rowBytes + x] == GC_PR_FGD) {

				((Pixel_8 *) binaryMask.data)[y * binaryMask.rowBytes + x] = mode;
			}
		}
	}
}

/*
 function which responsable for refining mask after main algorithm output
 sourceImage - ARGB buffer
 binaryMask - Planar8 buffer (output of letTheMagicBegin function)
 refineMask - Planar8 buffer (user selected region in "refine" mode)
 mode - 0 - erasing user selected regions from main function output mask using algorithm
        255 - adding user selected region on main function output mask using algorithm
 progress - algorithm state in percentages
 */
void letTheMagicRefine(const vImage_Buffer &sourceImage, vImage_Buffer &binaryMask,
					   vImage_Buffer &refineMask, int mode, int &progress) {

	progress = 0;
	auto weightsBufferSize = sizeof(float *) * sourceImage.width * sourceImage.height;

	GMM bgdGMM, fgdGMM;

	vImage_Buffer compIdxs = {malloc(sizeof(Pixel_8 *) * sourceImage.width * sourceImage.height),
							  sourceImage.height, sourceImage.width, sourceImage.width};

	int pathAmount = initGMMs(bgdGMM, fgdGMM, sourceImage, refineMask);
	if (pathAmount == -1) {
		return;
	}

	progress = 5;
	const double gamma = 50;
	const double lambda = 9 * gamma;
	const double beta = calcBeta(sourceImage);

	vImage_Buffer leftWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
							sourceImage.width};
	vImage_Buffer upleftWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
							  sourceImage.width};
	vImage_Buffer upWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
						  sourceImage.width};
	vImage_Buffer uprightWB = {malloc(weightsBufferSize), sourceImage.height, sourceImage.width,
							   sourceImage.width};

	calcNWeights(beta, gamma, sourceImage, leftWB, upleftWB, upWB, uprightWB);
	progress = 10;

	GCGraph<double> graph;
	assignGMMsComponents(bgdGMM, fgdGMM, compIdxs, sourceImage, refineMask);
	progress = 20;

	learnGMMs(compIdxs, bgdGMM, fgdGMM, sourceImage, refineMask);
	progress = 25;

	constructGCGraph(bgdGMM, fgdGMM, lambda, graph, sourceImage, refineMask, leftWB, upleftWB, upWB,
					 uprightWB);
	progress = 50;

	vImage_Buffer resultMask = {malloc(sizeof(uint8_t) * binaryMask.height * binaryMask.width),
								binaryMask.height, binaryMask.width, binaryMask.width};

	auto resultAmount = refineSegmentation(graph, resultMask, binaryMask, mode);


	if ((float) resultAmount / pathAmount > 0.5f) {
		refineInput(resultMask, binaryMask, mode);
	} else {
		returnInput(refineMask, binaryMask, mode);
	}
	progress = 100;

	free(compIdxs.data);
	compIdxs.data = nullptr;
	free(leftWB.data);
	leftWB.data = nullptr;
	free(upleftWB.data);
	upleftWB.data = nullptr;
	free(upWB.data);
	upWB.data = nullptr;
	free(uprightWB.data);
	uprightWB.data = nullptr;
	free(resultMask.data);
	resultMask.data = nullptr;
}

extern "C" {
JNIEXPORT jboolean JNICALL Java_com_picsart_studio_editor_brushhelper_CutOutMarker_letTheMagicBegin(
		JNIEnv *env, jobject thiz, jobject jSourceBitmap, jobject jInputMaskBitmap) {

	AndroidBitmapInfo sourceInfo;

	int result = AndroidBitmap_getInfo(env, jSourceBitmap, &sourceInfo);
	if (result) {
		return (jboolean) false;
	}

	AndroidBitmapInfo inputMaskInfo;

	result = AndroidBitmap_getInfo(env, jInputMaskBitmap, &inputMaskInfo);
	if (result) {
		return (jboolean) false;
	}

	Pixel_8 *sourcePixels = NULL;

	result = AndroidBitmap_lockPixels(env, jSourceBitmap, (void **) &sourcePixels);
	if (result) {
		return (jboolean) false;
	}

	Pixel_8 *inputMaskPixels = NULL;

	result = AndroidBitmap_lockPixels(env, jInputMaskBitmap, (void **) &inputMaskPixels);
	if (result) {
		return (jboolean) false;
	}

	vImage_Buffer sourceImage = {sourcePixels, sourceInfo.height, sourceInfo.width,
								 sourceInfo.stride};
	vImage_Buffer inputMaskImage = {inputMaskPixels, inputMaskInfo.height, inputMaskInfo.width,
									inputMaskInfo.stride};

	AndroidBitmap_unlockPixels(env, jSourceBitmap);
	AndroidBitmap_unlockPixels(env, jInputMaskBitmap);

	int progress;

	for (int i = 0; i < inputMaskImage.width * inputMaskImage.height; i++) {
		if (((Pixel_8 *) inputMaskImage.data)[i] != 0) {
			((Pixel_8 *) inputMaskImage.data)[i] = 255;
		}
	}

	bool found = letTheMagicBegin(sourceImage, inputMaskImage, progress);
	smoothBuffer(inputMaskImage, 100);

	return (jboolean) found;
}

void Java_com_picsart_studio_editor_brushhelper_CutOutMarker_letTheMagicRefine(JNIEnv *env,
																			   jobject thiz,
																			   jobject jSourceBitmap,
																			   jobject jInputMaskBitmap,
																			   jobject jRefineMaskBitmap,
																			   int mode) {

	AndroidBitmapInfo sourceInfo;

	int result = AndroidBitmap_getInfo(env, jSourceBitmap, &sourceInfo);
	if (result) {
		return;
	}

	AndroidBitmapInfo inputMaskInfo;

	result = AndroidBitmap_getInfo(env, jInputMaskBitmap, &inputMaskInfo);
	if (result) {
		return;
	}

	AndroidBitmapInfo refineMaskInfo;

	result = AndroidBitmap_getInfo(env, jRefineMaskBitmap, &refineMaskInfo);
	if (result) {
		return;
	}

	Pixel_8 *sourcePixels = NULL;

	result = AndroidBitmap_lockPixels(env, jSourceBitmap, (void **) &sourcePixels);
	if (result) {
		return;
	}

	Pixel_8 *inputMaskPixels = NULL;

	result = AndroidBitmap_lockPixels(env, jInputMaskBitmap, (void **) &inputMaskPixels);
	if (result) {
		return;
	}

	Pixel_8 *refineMaskPixels = NULL;

	result = AndroidBitmap_lockPixels(env, jRefineMaskBitmap, (void **) &refineMaskPixels);
	if (result) {
		return;
	}

	vImage_Buffer sourceImage = {sourcePixels, sourceInfo.height, sourceInfo.width,
								 sourceInfo.stride};
	vImage_Buffer inputMaskImage = {inputMaskPixels, inputMaskInfo.height, inputMaskInfo.width,
									inputMaskInfo.stride};
	vImage_Buffer refineMaskImage = {refineMaskPixels, refineMaskInfo.height, refineMaskInfo.width,
									 refineMaskInfo.stride};

	AndroidBitmap_unlockPixels(env, jSourceBitmap);
	AndroidBitmap_unlockPixels(env, jInputMaskBitmap);
	AndroidBitmap_unlockPixels(env, jRefineMaskBitmap);

	int progress;

	letTheMagicRefine(sourceImage, inputMaskImage, refineMaskImage, mode, progress);
}
}