#include "Kmeans.hpp"
#include <random>

#ifndef EFFECTS_USE_GCD
#define EFFECTS_USE_GCD 0
#endif


#if EFFECTS_USE_GCD
#include <dispatch/dispatch.h>
#endif

#define DISPATCH_QUEUE_PRIORITY_HIGH 2
#define DISPATCH_QUEUE_PRIORITY_DEFAULT 0
#define DISPATCH_QUEUE_PRIORITY_LOW (-2)
#define DISPATCH_QUEUE_PRIORITY_BACKGROUND INT16_MIN

#define CONVERGED      0
#define NOT_CONVERGED -1

void initialize(Pixel* centroids,int num_clusters,Pixel* dataPoints,int num_points)
{
    int dex;
    std::random_device rd;
    for(dex=0;dex<num_clusters;dex++)
    {

        std::mt19937 rng(rd());
        std::uniform_int_distribution<int> uni(0,num_points);
        auto random_index = uni(rng);

        centroids[dex].red = dataPoints[random_index].red;
        centroids[dex].green = dataPoints[random_index].green;
        centroids[dex].blue = dataPoints[random_index].blue;

    }
}

void resetData(int *data,int num_points)
{
    for(int i = 0; i < num_points; i++)
    {
        data[i]=-1;
    }
}


double calculateDistance(Pixel point1,Pixel point2)
{
    return (pow((point1.red-point2.red)*100,2) + pow((point1.green-point2.green)*100,2) + pow((point1.blue-point2.blue)*100,2));
}


void calculateNewCentroids(Pixel* points,int* data,Pixel* centroids,int num_clusters,int num_points)
{
    Pixel* newCentroids = (Pixel*)malloc(sizeof(Pixel)*num_clusters);
    int* population = (int*)malloc(sizeof(int)*num_clusters);
    int i;

    for(i = 0; i<num_clusters; i++)
    {
        population[i] = 0;
        newCentroids[i].red = 0;
        newCentroids[i].green = 0;
        newCentroids[i].blue = 0;

    }
    for(i = 0;i < num_points; i++)
    {
        population[data[i]]++;
        newCentroids[data[i]].red += points[i].red;
        newCentroids[data[i]].green += points[i].green;
        newCentroids[data[i]].blue += points[i].blue;

    }
    for(i = 0;i < num_clusters; i++)
    {
        if(population[i]!=0.0)
        {
            newCentroids[i].red /= population[i];
            newCentroids[i].green /= population[i];
            newCentroids[i].blue /= population[i];

        }
    }
    for(i = 0;i < num_clusters; i++)
    {
        centroids[i].red = newCentroids[i].red;
        centroids[i].green = newCentroids[i].green;
        centroids[i].blue = newCentroids[i].blue;

    }

    free(newCentroids);
    newCentroids = nullptr;
    free(population);
    population = nullptr;
}



int checkConvergence(int *former_clusters,int *latter_clusters,int num_points)
{
    for(int i = 0; i < num_points; i++)
        if(former_clusters[i] != latter_clusters[i])
            return NOT_CONVERGED;
    return CONVERGED;
}


struct PointsDistsContext{

    Pixel* data_points;
    Pixel* centroids;
    int    num_centroids;
    int*   clusters;
};

static void computeDists_Parallel(void* context, size_t iter){

    PointsDistsContext* cnt = (PointsDistsContext*)context;
    Pixel point = cnt->data_points[iter];

    int parentCluster = 0;
    double distance = 0;
    double minDistance = calculateDistance(point,cnt->centroids[0]);

    for(int i = 1; i < cnt->num_centroids; i++)
    {
        distance = calculateDistance(point,cnt->centroids[i]);
        if(minDistance >= distance)
        {
            parentCluster = i;
            minDistance = distance;
        }
    }
    cnt->clusters[iter] = parentCluster;

}


void kmeans(Pixel* imgpoints, int k, int num_points, int* labels)
{

    int job_done=0;

    Pixel* centroids;
    int  * formerLabels;
    bool firstLoop = true;

    formerLabels = (int*)malloc(sizeof(int)*num_points);
    centroids = (Pixel*)malloc(sizeof(Pixel)*k);

    //reseting and initializing to default behaviour
    initialize(centroids,k,imgpoints,num_points);
    resetData(formerLabels,num_points);
    resetData(labels,num_points);

    if(!job_done && !firstLoop){

        calculateNewCentroids(imgpoints,labels,centroids,k,num_points);
        if(checkConvergence(labels,formerLabels,num_points)==0){

            job_done = 1;
        }else{
            for(int i = 0;i < num_points; i++){
                    formerLabels[i] = labels[i];
                }
            }

            if(job_done == 1)
                return;

            }else{
                firstLoop = false;
                PointsDistsContext context;
                context.centroids = centroids;
                context.clusters = labels;
                context.data_points = imgpoints;
                context.num_centroids = k;

                #if EFFECTS_USE_GCD
                    dispatch_apply_f(num_points, dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), &context, computeDists_Parallel);
                #else
                #pragma omp parallel for
                    for (int iter = 0; iter < num_points; ++iter)
                        computeDists_Parallel(&context, iter);
                #endif

        }

    free(formerLabels);
    formerLabels = nullptr;
    free(centroids);
    centroids = nullptr;
    }